## Associate contours and nuclei by distance 
## algorithm hongrois: Kuhn-Munkres
import numpy as np
from math import sqrt, floor
from scipy.ndimage.morphology import distance_transform_edt
from munkres import Munkres
from skimage.measure import label, regionprops

def distance2DCenters( cent0, cent1, scaleXY ):
    if cent0 is None:
        return 0
    return sqrt( (cent0[0]-cent1[0])*(cent0[0]-cent1[0]) + (cent0[1]-cent1[1])*(cent0[1]-cent1[1]) )*scaleXY

def associateWindows(wlab, wbal, dlim, scaleXY):
    labels = np.unique(wlab[wlab>0])
    balels = np.unique(wbal[wbal>0])
    nbal = len(balels)
    nlab = len(labels)
    resbal = np.zeros_like(wbal)
    n = max(nbal, nlab)
    matrix = np.zeros((n, n))  ## should be square. 
    if n == 0:
        return resbal

    ## distance_transform is distance to closest background, so inverse
    dist2lab, nearest_coord = distance_transform_edt(wlab==0, return_indices=True)
    dillabels = np.zeros_like(wlab)
    ## check if within reasonnable distance
    dilate_mask = ((dist2lab*scaleXY*scaleXY) <= dlim)
    masked_nearest_label_coords = [ dind[dilate_mask] for dind in nearest_coord]
    nearest_labels = wlab[tuple(masked_nearest_label_coords)]
    dillabels[dilate_mask] = nearest_labels

    ## look for best fit
    for j, jlab in enumerate(balels):
        nei = dillabels[wbal==jlab]
        for i, ilab in enumerate(labels):
                count = np.sum(nei==ilab)
                if count > 0:
                    matrix[i][j] = 1.0/(count+1)
                else:
                    matrix[i][j] = dlim+1
    
    ## algorithm hongrois: Kuhn-Munkres
    dmat = np.copy(matrix)
    munk = Munkres()
    assoc = munk.compute(matrix)
    matrix = None
    munk = None
    for asso in assoc:
        if asso[0]<nlab and asso[1]<nbal:
            if (dmat[asso[0]][asso[1]] > 0) and (dmat[asso[0]][asso[1]]<dlim):
                ## associate
                resbal[wbal==balels[asso[1]]] = labels[asso[0]]

    return resbal
    
def associateLabWithLab(lab, bal, dlim, scaleXY):
    """ associate labels of img 2 to labels of img 1 """
    ##### do overlapping windows otherwise calcul distances too slow (to test)
    sizex = 1000
    sizey = 1000
    over = 50
    sizes = lab.shape

    posy = 0
    resbal = np.zeros(bal.shape, dtype="uint16")
    while posy < sizes[0]:
        posx = 0
        while posx < sizes[1]:
            windowbal = np.copy(bal[posy:(posy+sizey),posx:(posx+sizex)])
            windowlab = lab[posy:(posy+sizey),posx:(posx+sizex)]
            assobal = associateWindows( windowlab, windowbal, dlim, scaleXY )
            bal[posy:(posy+sizey),posx:(posx+sizex)][assobal>0] = 0
            resbal[posy:(posy+sizey),posx:(posx+sizex)][assobal>0] = assobal[assobal>0]
                
            posx += (sizex-over)        
        posy += (sizey-over)

    ## what is left in bal image has not been associated, add it to the result image as new label
    maxlab = np.max(lab)
    for l in np.unique(bal):
        if l > 0:
            resbal[bal==l] = maxlab+1
            maxlab = np.max(resbal)
 
    return resbal     

def associateNucleus(labs, dlimit=3, scaleXY=1):
    """ Associate each slice with previous slice """
    for i in range(len(labs)):
        if i > 1:
            rlab = associateLabWithLab( labs[i-1,], labs[i,], dlimit, scaleXY )
            labs[i,] = rlab
    return labs

def associateOverlap( labimg, labimgprev, threshold_overlap=0.25):
    nuclei_prop = regionprops( labimg, intensity_image=labimgprev )
    new_label = np.max(labimgprev) + 1
    #taken = []
    for nucprop in nuclei_prop:
        overlap = nucprop.image_intensity
        overlabs, counts = np.unique(overlap, return_counts=True)
        if 0 in overlabs:
            zero = np.where(overlabs==0)
            zero = zero[0]
            overlabs = [over for i, over in enumerate(overlabs) if i != zero]
            counts = [over for i, over in enumerate(counts) if i != zero]
        
        done = False
        if len(counts)>0:
            maxlabind = np.argmax( counts )
            #maxlabind = maxlabind[0]
            if counts[maxlabind]/nucprop.area > threshold_overlap:
                ## overlap between the two labels, associate
                labimg[nucprop.bbox[0]:nucprop.bbox[2], nucprop.bbox[1]:nucprop.bbox[3]][nucprop.image] = overlabs[maxlabind]
                done = True

        if not done:
            # no match found, new label
            labimg[nucprop.bbox[0]:nucprop.bbox[2], nucprop.bbox[1]:nucprop.bbox[3]][nucprop.image] = new_label
            new_label = new_label + 1
            #taken.append(maxlabind)
    return labimg


def associateNucleusOverlap(labs, threshold_overlap):
    """ Associate each slice with previous slice based on IOU """
    for i, lab in enumerate(labs):
        if i >= 1:
            plab = associateOverlap( lab, plab, threshold_overlap )
        else:
            # first slice
            plab = lab
        labs[i,] = plab
    return labs 


def associate_objects(pop, wnuc, wcells, dlim, scaleXY, scaleZ):
    ## algorithm hongrois: Kuhn-Munkres
    ncells = len(wcells)
    nnuclei = len(wnuc)
    n = max(ncells, nnuclei)
    #print(str(ncells)+" "+str(nnuclei)+" "+str(n))

    matrix = np.zeros((n,n))  ## should be square.
    row = 0
    for i, cell in enumerate(wcells):
        for j, nuc in enumerate(wnuc):
            matrix[i][j] = pop.distanceNucleusToCell( nuc, cell, scaleXY, scaleZ )  ## put more weights to XY distance than z

    munk = Munkres()
    dmat = np.copy(matrix)
    assoc = munk.compute(matrix)
    munk = None
    associated = []
    associatedNuc = []
    #acells = []
    ## assoc contient indices. If indices > nnuclei or ncells, mean non associated
    for asso in assoc:
        if asso[0] < ncells and asso[1] < nnuclei:
            if dmat[asso[0]][asso[1]] < dlim:
                ## associate them
                pop.associateNucleusAndRelabel( nucleus=wnuc[asso[1]], cell=wcells[asso[0]] )
                associated.append(asso[0])
                associatedNuc.append(asso[1])
       
    dmat = None
    return associated, associatedNuc   

        
def associate_nucleiToCell(pop, imgsizes, dlim=20, scaleXY=1, scaleZ=1, pbar=None):
    """ Need scaling for non isotropic distances """
    ##### do overlapping windows otherwise association too slow (too big matrix)
    sizex = floor(180/scaleXY)
    sizey = floor(180/scaleXY)
    over = floor(40/scaleXY)
    margin = floor(5/scaleXY)

    posy = 0
    #fullcells = []  ## cells with associated nuclei
    nuclei = pop.nuclei.values()
    cells = pop.cells.values()
    if pbar is not None:
        pbar.total = imgsizes[0]
    while posy < imgsizes[0]:
        if pbar is not None:
            pbar.update(posy)
        posx = 0
        while posx < imgsizes[1]:
            border = (posy, posx, posy+sizey, posx+sizex)
            
            winuclei = []
            leftnuclei = []
            #print("newcnts "+str(len(newcnts)))
            for nuc in nuclei:
                    if nuc.insideBorderCenter(border):
                        winuclei.append(nuc)
                    else:
                        leftnuclei.append(nuc)
            nuclei = leftnuclei
            
            wincells = []
            border = (posy-margin, posx-margin, posy+sizey+margin, posx+sizex+margin)
            leftcells = []
            for cell in cells:
                    if cell.insideBorderCenter(border):
                        wincells.append(cell)
                    else:
                        leftcells.append(cell)
            
            if len(winuclei)>0 and len(wincells)>0:
                associated, associatedNuc = associate_objects(pop, winuclei, wincells, dlim, scaleXY, scaleZ)
                #fullcells = fullcells + acells
                
                for i in range(len(wincells)):
                    if i not in associated:
                        leftcells.append(wincells[i])
                
                for i in range(len(winuclei)):
                    if i not in associatedNuc:
                        nuclei.append(winuclei[i])
                        
            cells = leftcells
                
            posx += (sizex-over)        
        posy += (sizey-over)
 
    print("Unassociated nuclei left "+str(len(nuclei)))
    pop.relabelUnassociatedNuclei(nuclei)



import numpy as np
import os, time
from math import floor
import csv
from skimage.exposure import adjust_gamma
import skimage.metrics as imetrics
import fish_feats.Utils as ut
from fish_feats.CellObjects import Population

try:
    from fish_feats.SegmentObj import local_max_proj, prepJunctions, segmentJunctions
    from fish_feats.RNASpots import RNASpots
except ImportError:
    print("Module missing")
    
####### Z map functions
def score_each_z( img3d, projimg ):
    """ Compare each z image to the projected image """
    best_score = [0,0]
    for z, zslice in enumerate(img3d):
        score = imetrics.normalized_mutual_information(zslice, projimg)
        if best_score[0] < score:
            best_score[0] = score
            best_score[1] = z
    return best_score[1]

def process_x(x, step, projimg, img, winsize):
    """ Process all the image and construct the zmap """
    xstep = step

    if (x+xstep) > projimg.shape[0]:
        xstep = projimg.shape[0]-x
    zmap_x = np.zeros((xstep,projimg.shape[1]), "uint8")
    xmin = x-int(winsize/2)
    if xmin < 0:
        xmin = 0
    xmax = xmin+winsize
    if xmax > projimg.shape[0]:
        xmax = projimg.shape[0]
        xmin = projimg.shape[0]-winsize
    for y in range(0,projimg.shape[1],step):
        ymin = y-int(winsize/2)
        if ymin < 0:
            ymin = 0
        ymax = ymin+winsize
        if ymax > projimg.shape[0]:
            ymax = projimg.shape[0]
            ymin = projimg.shape[0]-winsize

        img3d = img[:,xmin:xmax,ymin:ymax]
        proj = projimg[xmin:xmax,ymin:ymax]
        bestz = score_each_z(img3d, proj)
        zmap_x[:,y:(y+step)] = bestz
    return zmap_x

class MainImage:
    """ Handle the main (original) image (scaling, channels), interface with other analysis """

    def __init__(self, talkative =True):
        self.scaleXY = 1.0
        self.scaleZ = 1.0
        self.junchan = 0
        self.nucchan = 0
        self.nbchannels = 1
        self.verbose = talkative
        self.zdirection = -1
        self.pop = None
        self.nucstain = None
        self.junstain = None
        self.junmask = None
        self.nucmask = None
        self.model = None
        self.rnas = {}
        self.imshape = (0,0,0)
        self.imagename = None
        self.imagedir = None

    def set_imagename(self, imagename):
        """ Set the name and path of the image """
        self.imagename, self.imagedir, self.resdir = ut.extract_names( imagename )
        self.pop = Population( imageshape=(0,0,0) )

    def set_scales( self, scaleZ, scaleXY ):
        """ Set the scales for the image """
        self.scaleXY = scaleXY
        self.scaleZ = scaleZ

    def get_image_path( self ):
        """ Returns the full image name and directory """
        if self.imagename is None:
            return ""
        if self.imagedir is None:
            return ""
        return os.path.join(self.imagedir, self.imagename)

    def set_image_path(self, imgpath=None):
        """ Set the image name and result directory """
        self.imagename, self.imagedir, self.resdir = ut.extract_names( imgpath )

    def crop_name( self ):
        """ build default crop name """
        return os.path.join( self.imagedir, self.imagename+"_crop.tif" )

    def set_image( self, images ):
        """ Set the main image (all the channels) from list of images (already opened) """
        self.image = images
        self.nbchannels = len(images)
        self.imshape = images[0].shape
        self.pop = Population(imageshape=self.imshape)

    def open_image(self, filename):
        self.image, self.scaleXY, self.scaleZ, names = ut.open_image(filename, verbose=self.verbose)
        self.imagename, self.imagedir, self.resdir = ut.extract_names( filename )
        self.image = ut.arrange_dims(self.image, verbose=self.verbose)
        self.imshape = self.image[0,].shape
        self.nbchannels = self.image.shape[0]
        self.pop = Population(imageshape=self.imshape)

    def image_2dshape(self):
        if len(self.imshape) == 3:
            return self.imshape[1:]
        return self.imshape
    
    def get_image_shape(self, in2d=True):
        if in2d:
            if self.junmask is not None:
                return self.junmask.shape
            else:
                if self.junstain is None:
                    return None
                return self.junstain.shape[1:]
        else:
            return self.junstain.shape

    def get_coordinates(self, pos):
        """ Get pixel coordinates from tuple of position """
        ind = int(len(pos)==3)
        return ( int(pos[ind]/self.scaleXY), int(pos[ind+1]/self.scaleXY) )

    def free_channel(self):
        for i in range(self.nbchannels):
            if i != self.junchan:
                if i != self.nucchan:
                    return i
        return 0

    def potential_rnas(self):
        """ Return the list of potentially RNA channels (no junctions nor nuclei) """
        channels = []
        for chan in range(self.nbchannels):
            if (chan != self.junchan) and (chan != self.nucchan):
                channels.append(chan)
        return channels

    def get_channel(self, chan):
        #return zarr.array(self.image[chan])
        return self.image[chan]

    def build_filename(self, endname):
        return os.path.join(self.resdir, self.imagename+endname)

    def get_filename(self, endname, ifexist=False):
        """ Return filename if its exist and ifexist True """
        resfile = os.path.join( self.resdir, self.imagename+endname )
        if ifexist:
            if not os.path.exists(resfile):
                return ""
        return resfile

    def save_zcells(self, filename):
        """ Save a file with the list of cells and corresponding Z position """
        print("Saving Z positions of cells in file "+str(filename))
        zdict = self.pop.getCellsZDict()
        if zdict == []:
            ## calculate the z positions if it was not yet done
            self.popFromJunctions()
            zdict = self.pop.getCellsZDict()
        ut.write_dict(filename, zdict)

    def save_image( self, img, imagename=None, hasZ=False, endname=None, imtype="uint16" ):
        if hasZ:
            scaZ = self.scaleZ
        else:
            scaZ = -1
        if imagename is None:
            imagename = self.build_filename(endname)
        if int(np.max( img, axis=None )) >= 65535: 
            imtype="float32"
        ut.writeTif( img, imagename, scaleXY=self.scaleXY, scaleZ=scaZ, imtype=imtype)
        ut.show_info( "Image "+str(imagename)+" saved" )

    def rnacount_filename(self, ifexist=True):
        resfile = os.path.join(self.resdir, self.imagename+"_RNACounts.csv")
        if ifexist:
            if not os.path.exists(resfile):
                return ""
        return resfile

    def save_measures( self, results, endname, keepprevious=True ):
        """ Save the table of measurement to file. If the file already exists, append the new measures if keepprevious """
        if (results is None) or (len(results) == 0):
            return
        resfile = os.path.join( self.resdir, self.imagename+endname )
        if keepprevious and os.path.exists( resfile ):
            results = self.load_measure_file( resfile, results )
        ut.write_dict(resfile, results)
        ut.show_info("Measures saved in file "+str(resfile))

    def save_results( self ):
        """ Save the table of all results to file """
        resfile = self.get_filename( endname="_results.csv", ifexist=False )
        results = self.pop.getResults( self.scaleXY, self.scaleZ )
        if (results is None) or (len(results) == 0):
            return
        ut.write_dict( resfile, results )
        ut.show_info("Results saved in file "+str(resfile))

    def load_results_file( self, resfile ):
        """ Load already measured table if the corresponding file exists """
        rows = {}
        hasZ = False
        with open( resfile, 'r' ) as infile:
            csvreader = csv.DictReader(infile)
            for row in csvreader:
                if "CellLabel" not in row.keys():
                    continue
                if row["CellLabel"] == "":
                    continue
                try:
                    rows[ int(float(row["CellLabel"])) ] = row
                    if ("PosZPixel" in row.keys()) and (int(float(row["PosZPixel"])) > 0 ):
                        hasZ = True
                except:
                    continue
        return rows, hasZ

    def load_from_results( self, resfile ):
        """ Load results from file and create cells"""
        results, hasz = self.load_results_file( resfile )
        if self.pop is None:
            self.pop = Population( imageshape=self.imshape )
        self.pop.createCellsFromResults( results, direction=self.zdirection, talkative=self.verbose )
    
    def load_measure_file( self, resfile, results ):
        """ Load already measured table if the corresponding file exists """
        with open( resfile, 'r' ) as infile:
            columns = results[0].keys()
            csvreader = csv.DictReader(infile)
            rows = []
            for row in csvreader:
                rows.append( row )
            previous_measures = set(rows[0].keys()) - set(columns)
            print(previous_measures)
            if len( previous_measures) > 0:
                print( "Appending previous measures value in the table" )
                results = self.pop.measureAppendPrevious( results, rows, previous_measures )
            return results


    def save_filename(self):
        return os.path.join(self.resdir, self.imagename+".cfg")

    def load_image(self, filename):
        loadedimg, scaleXYm, scaleZm, names = ut.open_image(filename, verbose=self.verbose)
        return loadedimg

    ################# Junctions

    def junction_filename(self, dim=2, ifexist=False):
        end = "_cells2D.tif"
        if dim==3:
            end = "_cells3D.tif"
        return self.get_filename(end, ifexist)
    
    def zcell_filename(self, ifexist=False):
        end = "_cellsZ.csv"
        return self.get_filename(end, ifexist)
    
    def junction_projection_filename(self, ifexist=False):
        end = "_junction_projection.tif"
        return self.get_filename(end, ifexist)

    def getJunctionsProjection(self):
        """ Return the projection of the junction staining """
        projfilename = self.junction_projection_filename(ifexist=True)
        if projfilename != "":
            proj = self.load_image(projfilename)
            return proj

        ## No projection file found, have to redo the projection
        if self.junstain is None:
            return self.prepare_segmentation_junctions()
        
        ## Something failed somewhere
        print("Something failed in calculating/loading the junction projection")
        return None
    
    def getJunctionsImage3D( self, full=True, thick=1 ):
        """ Create a 3D image containing the cell labels at their 3D position"""
        if self.junstain is None:
            self.junstain = self.image[self.junchan,]
        return self.pop.drawCellsJunc3D( self.junstain, full, thick )

    def loadCellsFromSegmentation( self, junfilename ):
        """ Load only the segmentation mask """
        self.junmask, scaleXYm, scaleZm, names = ut.open_image(junfilename, verbose=self.verbose)
        if (self.junmask.dtype == np.float32) or (self.junmask.dtype == np.float64 ):
            self.junmask = self.junmask.astype(np.uint32)
        self.pop = Population(imageshape=self.imshape)
        self.pop.createCellsFromMask(self.junmask, None, None, find_z=False, zmap=None, cells_direction=self.zdirection, talkative=self.verbose)

        ## if cellsZ file exists, load it
        zcellfilename = self.zcell_filename(ifexist=True)
        if zcellfilename != "":
            zcellsDict = ut.load_dictlist(zcellfilename, verbose=self.verbose)
            self.pop.updateCellsZPosFromDictList( zcellsDict )     


    def load_segmentation( self, junfilename ):
        """ Load the segmentation and the sepatated staining if necesary"""
        self.junmask, scaleXYm, scaleZm, names = ut.open_image(junfilename, verbose=self.verbose)
        if (self.junmask.dtype == np.float32) or (self.junmask.dtype == np.float64 ):
            self.junmask = self.junmask.astype(np.uint32)
        if self.junstain is None:
            if self.should_separate():
                separated_junctionsfile = self.separated_junctions_filename(ifexist=True)
                separated_nucleifile = self.separated_nuclei_filename(ifexist=True)
                if (separated_junctionsfile != "") and (separated_nucleifile != ""):
                    self.load_separated_staining( separated_junctionsfile, separated_nucleifile )
                else:
                    print("Should separate the junctions and nuclei staining first")
            else:
                self.junstain = self.image[self.junchan,]
    
    def separated_nuclei_filename(self, ifexist=False):
        end = "_nucleiStaining.tif"
        return self.get_filename(end, ifexist)
    
    def separated_junctions_filename(self, ifexist=False):
        end = "_junctionsStaining.tif"
        return self.get_filename(end, ifexist)
        
    def load_separated_staining(self, junfile, nucfile ):
        """ Load the separated staining from files """
        self.junstain, scaleXYm, scaleZm, names = ut.open_image(junfile, verbose=self.verbose)
        self.nucstain, scaleXYm, scaleZm, names = ut.open_image(nucfile, verbose=self.verbose)

    def separate_junctions_nuclei( self, wth_radxy=4, wth_radz=1, rmoutlier_radxy=5, rmoutlier_radz=1, rmoutlier_threshold=40, smoothnucxy=2, smoothnucz=2 ):
        if self.junchan!=self.nucchan:
            ut.show_info("Not in the same channel, no separation necessary")
            self.nucstain = np.copy(self.image[self.nucchan,])
            self.junstain = np.copy(self.image[self.junchan,])
            return
        from fish_feats.Separe import separateNucleiJunc, smoothNuclei
        ## Try to separate junction and nucleus that are in the same staining
        nucimage = np.copy(self.image[self.nucchan,])
        self.nucstain, self.junstain = separateNucleiJunc(nucimage, outrz=rmoutlier_radz, outrxy=rmoutlier_radxy, threshold=rmoutlier_threshold, edge=5, space=50, zhatrad=wth_radz, hatrad=wth_radxy)
        self.nucstain = smoothNuclei(self.nucstain, radxy=smoothnucxy, radz=smoothnucz)
    
    def separate_with_sepanet( self, model_dir ):
        from fish_feats.Separe import sepanet
        bothimage = np.copy(self.image[self.nucchan,])
        self.junstain, self.nucstain = sepanet( bothimage, model_dir )

    def should_separate( self ):
        if self.junchan==self.nucchan:
            ## both channels are absent
            if self.junchan is None:
                return False
            # separation has not been done yet
            if self.junstain is None:
                return True
        return False

    def check_separation(self):
        """ Check if should separate, do it if yes """
        if self.junchan==self.nucchan:
            # separation has not been done yet
            if self.junstain is None:
                self.separate_junctions_nuclei()
            return True
        return False
    
    def preprocess_junction_removebg(self, radius):
        """ Remove background in junction staining """
        if self.junstain is None:
            self.check_separation()
        from fish_feats.SegmentObj import preprocessRemoveBg
        radXY = radius
        radZ = floor(self.scaleZ/self.scaleXY*radius)
        self.junstain = preprocessRemoveBg(self.junstain, radXY, radZ)
    
    def preprocess_junction2D_removebg(self, img, radius):
        """ Remove background in projed junction staining """
        from fish_feats.SegmentObj import preprocessRemoveBg2D
        return preprocessRemoveBg2D(img, radius)
    
    def preprocess_junction2D_tophat(self, img, radius):
        """ Clean with top hat in junction staining """
        from fish_feats.Separe import topHat2D
        return topHat2D(img, radius)
    
    def preprocess_junction_tophat(self, radius):
        """ Clean with top hat in junction staining """
        if self.junstain is None:
            self.check_separation()
        from fish_feats.Separe import topHat
        radXY = radius
        self.junstain = topHat(self.junstain, radXY)

    def prepare_segmentation_junctions( self, projxy=40, projsmooth=3, do_clahe=True, clahe_grid=10 ):
        """ Check if should separate. Do local proj and normalize """
        if not self.check_separation():
            self.junstain = self.image[self.junchan,]

        start_time = time.time()
        projzo1 = local_max_proj( self.junstain, largexy=projxy, smooth=projsmooth )
        roijunc = prepJunctions( projzo1, do_clahe=do_clahe, clahe_grid=clahe_grid )
        if self.verbose:
            print("Local projection finished in {:.3f}".format((time.time()-start_time)/60)+" min")
        return roijunc
    
    def prepare_junctions(self):
        if self.junstain is None:
            self.check_separation()
    
    def do_segmentation_junctions( self, methodJunctions, roijunc, cell_diameter=20, chunking=1000 ):
        """ Junctions ZO1 segmentation """
        start_time = time.time()
        if methodJunctions == "Empty":
            self.junmask = np.zeros( self.get_image_shape(in2d=True), dtype="uint16" )
            return
        self.junmask = segmentJunctions( roijunc, methodJunctions, self.scaleXY, self.imagedir, self.imagename, diameter=cell_diameter, chunking=chunking, verbose=self.verbose ) 
        self.junmask[self.junmask==1] = np.max(self.junmask)+1 ## no label 1
        ut.show_info("Segmentation finished in {:.3f}".format((time.time()-start_time)/60)+" min")
    
   
    def get_tubeness_image(self, junc2d):
        from fish_feats.SegmentObj import image_tubeness
        return image_tubeness( junc2d )

    def measure_junctions_header(self):
        return self.pop.measureCellHeading()

    def measure_junctions(self):
        """ Measure the cells, keeping previous measures/features """
        if self.pop is None:
            self.popFromJunctions()

        results = self.pop.getResults( self.scaleXY, self.scaleZ )
        return results
    
    def get_cell_fromcoord( self, coord ):
        """ Get the cell label from coordinates """
        if self.pop is None:
            return None
        return self.pop.getCellFromCoord( coord )

    def popFromJunctions(self, proj=None, zpos=False ):
        """ create cells from the current masked image of junctions """
        #start_time = time.time()
        if self.pop is None:
            self.pop = Population(imageshape=self.imshape)

        if (self.junmask is None) or (self.junstain is None):
            ut.show_warning( "Mask was not loaded or junction and nuclei signals must be separated before")

        print("Creating cells from junctions")
    
        ## Load measures file if exists
        measurefile = self.get_filename( endname="_results.csv", ifexist=True )
        results = None
        hasz = False
        if measurefile != "":
            print( "Load previous results from file "+measurefile )
            results, hasz = self.load_results_file( measurefile )

        ## Calculates the Z position if necessary
        zmap = None
        if zpos and not hasz:
            if proj is None:
                projfilename = self.junction_projection_filename(ifexist=True)
                if projfilename != "":
                    proj = self.load_image(projfilename)
            if proj is not None:
                print("Calculating the z-map to place cells in 3D")
                zmap = self.calculate_zmap(proj, step_size=200, window_size=250)
       
        self.pop.createCellsFromMask(self.junmask, self.junstain, results, find_z=zpos, zmap = zmap, cells_direction=self.zdirection, talkative=self.verbose)
        ut.show_info("Created "+str(len(self.pop.cells))+" cells")
        return 1

    def cellsHaveZPos( self ):
        """ Check if the cells have Z position """
        if self.pop is None:
            return False
        return self.pop.cellsHaveZPos()

    def updateCellsZPos( self, step_size=100, window_size=150, save=False ):
        """ Calculate the zmap and replace the cells accordingly """
        proj = self.getJunctionsProjection()
        zmap = self.calculate_zmap(proj, step_size=step_size, window_size=window_size)
        self.pop.updateCellsZPosFromMap(zmap)
        if save:
            filename = self.get_filename( endname="_zmap.tif" )
            self.save_image( zmap, filename, hasZ=False, imtype="uint8" )
        
    def updateCellZPos(self, cell_label, zpos, img=None):
        """ Update the position of the given cell """
        self.pop.updateCellZPos(cell_label, zpos, img, self.scaleZ)

    def updateCellsZPosFromList(self, zlist):
        return None
    
    def hasCells(self):
        if self.pop is None:
            return False
        if self.junmask is None:
            return False
        if self.pop.imgcell is None:
            return False
        return True
    
    def nbCells(self):
        return len(self.pop.cells)
    
    def nbNuclei(self):
        return len(self.pop.nuclei)

    ################# Nuclei

    def nuclei_filename(self, ifexist=False):
        end = "_nuclei.tif"
        resfile = os.path.join(self.resdir, self.imagename+end)
        if ifexist:
            if not os.path.exists(resfile):
                return ""
        return resfile
    
    def load_segmentation_nuclei( self, filename, load_stain=True ):
        """ Load the results of nuclei segmentation from file"""
        self.nucmask, scaleXYm, scaleZm, names = ut.open_image(filename, verbose=self.verbose)
        if (self.nucmask.dtype == np.float32) or (self.nucmask.dtype == np.float64):
            self.nucmask = self.nucmask.astype(np.uint32)
        self.pop.setNucleiImage(self.nucmask)
        if load_stain and (self.nucstain is None):
            self.nucstain = self.image[self.nucchan,]
    
    def prepare_segmentation_nuclei( self ):
        if not self.check_separation():
            self.nucstain = self.image[self.nucchan,]
        else:
            if self.nucstain is None:
                self.separate_junctions_nuclei()

    def preprocess_nuclei_removebg(self, radius):
        if self.nucstain is None:
            self.prepare_segmentation_nuclei()
        from fish_feats.SegmentObj import preprocessRemoveBg
        radXY = radius
        radZ = floor(self.scaleZ/self.scaleXY*radius)
        self.nucstain = preprocessRemoveBg(self.nucstain, radXY, radZ)

    def prepare_nuclei(self):
        if self.nucstain is None:
            self.prepare_segmentation_nuclei()

    def preprocess_nuclei_median(self, radius):
        if self.nucstain is None:
            self.prepare_segmentation_nuclei()
        from fish_feats.SegmentObj import preprocessNucleiMedianFilter
        radXY = radius
        radZ = floor(self.scaleZ/self.scaleXY*radius)
        if radZ == 0:
            radZ = 1
        self.nucstain = preprocessNucleiMedianFilter(self.nucstain, radXY, radZ)
        
    # cellpose3D
    def do_segmentation_cellpose(self, diameter, threshold, resample=True, in3D=True, stitch_threshold=0.25, dask=False, chunk=1000 ):
        ut.show_info("Segmenting nuclei with CellPose3D")
        if self.model is None:
            from fish_feats.SegmentObj import initialize_cellpose
            self.model = initialize_cellpose()
        
        from fish_feats.SegmentObj import prepNuclei, getNuclei_cellpose3D
        treatedNuclei = prepNuclei(self.nucstain)  ## normalize the image
        treatedNuclei = np.expand_dims(treatedNuclei, axis=0)  ## add channel dimension
        self.nucmask = getNuclei_cellpose3D( self.model, treatedNuclei, 
                    diameter=diameter, 
                    scaleXY=self.scaleXY, scaleZ=self.scaleZ,
                    threshold=threshold,
                    flow_threshold=0.4,
                    resample=resample,
                    in3D=in3D,
                    stitch_threshold=stitch_threshold,
                    dask=dask,
                    chunk=chunk,
                    verbose=self.verbose )    
        self.nucmask[self.nucmask>0] = self.nucmask[self.nucmask>0] + 1
        if self.pop is not None:
            self.pop.setNucleiImage(self.nucmask)
    
    # stardist2D+association 3D
    def do_segmentation_stardist(self, threshold, overlap, assoMethod, associationlim, threshold_overlap):
        ut.show_info("Segmenting nuclei with Stardist2D+association3D")
        from fish_feats.SegmentObj import prepNuclei, getNuclei_stardist2DAsso3D
        treatedNuclei = prepNuclei(self.nucstain)  ## normalize the image
        self.nucmask = getNuclei_stardist2DAsso3D( treatedNuclei, self.scaleXY,
                            proba=threshold, 
                            overlap=overlap, 
                            assoMode = assoMethod,
                            assolim=associationlim,
                            threshold_overlap=threshold_overlap,
                            verbose=self.verbose )
        if self.nucmask is None:
            return
        self.nucmask[self.nucmask>0] = self.nucmask[self.nucmask>0] + 1
        if self.pop is not None:
            self.pop.setNucleiImage(self.nucmask)

    def get_segmented_nuclei(self):
        """ Mark points that are inside segmented nuclei """
        if self.pop is None:
            return None
        if self.pop.imgnuc is None:
            return None 
        return self.pop.imgnuc

    def hasNuclei(self):
        if self.pop is None:
            return False
        if self.pop.imgnuc is None:
            return False
        return True

    def filterNuclei( self, minimum_volume, keep_ifnz=2 ):
        if self.pop is None:
            self.pop = Population(imageshape=self.imshape)
        if self.pop.imgnuc is None:
            self.pop.imgnuc = self.nucmask
        self.pop.filterNuclei( minimum_volume, keep_ifnz, self.scaleXY*self.scaleXY*self.scaleZ, self.verbose )

    def popNucleiFromMask(self, associate=True):
        """ Load the nuclei and associate them to cells """
        self.pop.setNucleiImage(self.nucmask)
        self.pop.createNucleiFromMask( associate=associate, verbose=self.verbose, scaleXY=self.scaleXY, scaleZ=self.scaleZ )

    def measure_counts(self):
        rnachannels = [i for i in self.rnas.keys()]
        methods = [self.rnas[rnac].countName for rnac in rnachannels]
        self.pop.measureCellsCount( rnachannels, methods )

    def measure_count( self, rnachanel, method ):
        self.pop.measureCellsCount( [rnachanel], [method] )

    def get_counts( self ):
        """ Get the counts of the cells """
        if self.pop is None:
            return None
        return self.pop.getCounts()

    def get_measure_name(self, rnac):
        """ Return the name of the counting method of rna channel """
        return self.rnas[rnac].countName
    
    def image_count_from_table(self, countname):
        """ Create the image of cell RNA counts from current table """
        countimg = np.zeros(self.get_image_shape(in2d=True), np.uint16)
        self.pop.drawCountsInCells(countimg, countname)
        return countimg

    ### RNA hierarchical analysis

    def measure_onlycounts(self, rnachannels):
        methods = [self.rnas[rnac].countName for rnac in rnachannels]
        return self.pop.measureOnlyCellsCount(rnachannels, methods, self.scaleXY, self.scaleZ)

    def set_cells(self, clusterimg, clusters, celllabels):
        """ Put each cell value to its cluster """
        self.pop.drawCellsValues(clusterimg, clusters, celllabels)
    
    
    ###############  RNA

    def rna_filename(self, chan=0, how=".csv", ifexist=False):
        """ Default name of RNA files, can be an image or a csv """
        if how == ".tif":
            end = "_RNA"+str(chan)+".tif"
        elif how == ".csv":
            end = "_RNA"+str(chan)+".csv"
        return self.get_filename(end, ifexist)
    
    def load_rnafile( self, rnafilename, rnac, topop=True ):
        """ Load RNA spots for given channel from file """
        pop = None
        if topop:
            pop = self.pop
        rnaspot = RNASpots( rnac, verbose=self.verbose )
        self.rnas[rnac] = rnaspot
        ext = os.path.splitext(rnafilename)[1]
        # load from an image
        if ext == ".tif":
            rnaspots, scaleXYm, scaleZm, names = ut.open_image(rnafilename, verbose=self.verbose)
            rnaspot.update_spotsFromImage(rnaspots, methodName="Load", pop=pop)
        # load from a table file
        elif ext == ".csv":
            rnaspotDict = ut.load_dictlist(rnafilename, verbose=self.verbose)
            rnaspot.update_spotsFromDict(rnaspotDict, methodName="Load", pop=pop)
        ut.show_info("RNA spots loaded from file "+str(rnafilename))
    
    def get_rna_threshold(self, rnachan):
        if self.rnas.get(rnachan) is not None:
            if self.rnas[rnachan].threshold is None:
                return 0
            return self.rnas[rnachan].threshold
        return 0

    def find_rna(self, rnachan, spotZRadius, spotXYRadius, rmExtremeZ, threshold=None):
        if self.rnas.get(rnachan) is None:
            rnaspot = RNASpots(rnachan, verbose=self.verbose)
            self.rnas[rnachan] = rnaspot
        else:
            rnaspot = self.rnas[rnachan]
            rnaspot.reset_spots()
        rnaimg = self.image[rnachan,:,:,:]
        rnaspot.detect_spots_withbigfish(rnaimg, scaleZ=self.scaleZ, scaleXY=self.scaleXY, spotZRadius=spotZRadius, spotXYRadius=spotXYRadius, rmextr=rmExtremeZ, threshold=threshold)
    
    def find_rnas_in_image(self, rnaimg, chanlist, spotZRadius, spotXYRadius, threshold=None):
        chanlist = str(chanlist)
        if self.rnas.get(chanlist) is None:
            rnaspot = RNASpots(chanlist, verbose=self.verbose)
            self.rnas[chanlist] = rnaspot
        else:
            rnaspot = self.rnas[chanlist]
            rnaspot.reset_spots()
        rnaspot.detect_spots_withbigfish(rnaimg, scaleZ=self.scaleZ, scaleXY=self.scaleXY, spotZRadius=spotZRadius, spotXYRadius=spotXYRadius, rmextr=False, threshold=threshold)
        return rnaspot.get_points()

    def find_blobs_in_image(self, rnaimg, chanlist, minSigma, threshold):
        """ Detect 3D blobs (local maxima with DoG filtering) """
        chanlist = str(chanlist)
        if self.rnas.get(chanlist) is None:
            rnaspot = RNASpots(chanlist, verbose=self.verbose)
            self.rnas[chanlist] = rnaspot
        else:
            rnaspot = self.rnas[chanlist]
            rnaspot.reset_spots()
        rnaspot.detect_spots_withblobs(rnaimg, min_sigma=minSigma, threshold=threshold)
        return rnaspot.get_points()


    def assign_fromclosestrna(self, rnachan, method, limDist, nAbove, refrnas, nclosest):
        ut.show_info("Assign RNAs channel "+str(rnachan)+" with method "+method)
        rnaspot = self.rnas[rnachan]
        rnaclouds = [self.rnas[refrna] for refrna in refrnas]
        rnaspot.assign_fromcloud(self.pop, clouds=rnaclouds, 
                         scaleXY=self.scaleXY,
                         scaleZ=self.scaleZ,
                         distanceLimit=limDist, 
                         nclosest=nclosest,
                        method=method )

    def assign_rna(self, rnachan, method, limitDistAsso, nAbove):
        """ Assign the RNA spots to the cells with the defined parameters """
        angular_step = 4
        ut.show_info("Assign RNAs channel "+str(rnachan)+" with method "+method)
        if not self.hasCells():
            if method == "Projection" or method=="MixProjClosest" or method=="ClosestNucleus" or method== "Hull":
                ut.show_info("Cells not segmented yet, segment them before to use "+method+" method")
                return
        if method == "Hull" and self.pop.hullinited == 0:
            self.pop.initializeHull(angular_step)
        
        rnaspot = self.rnas[rnachan]
        rnaspot.assign_spots(self.pop, method=method,
                         distanceLimit=limitDistAsso, above=nAbove,
                         angular_step = angular_step,
                         scaleXY=self.scaleXY,
                         scaleZ=self.scaleZ)

    def draw_spots(self, spots, labels, chan):
        self.rnas[chan].update_spotsFromPoints(spots, labels, None)
        return self.rnas[chan].draw_spots3D(self.imshape, size=2)

    def save_spots(self, spots, props, chan, rnafile=None):
        """ Save the spots information to csv file """
        if rnafile is None:
            rnafile = self.rna_filename(chan=chan, how=".csv")
        ext = os.path.splitext(rnafile)[1]
        if ext == ".tif":
            imgspots = self.draw_spots( spots, props["label"], chan )
            self.save_image(imgspots, rnafile, hasZ=True )
            del imgspots
        if ext == ".csv":
            resdict = dict()
            resdict["X"] = spots[:,1]
            resdict["Y"] = spots[:,2]
            resdict["Z"] = spots[:,0]
            resdict["Label"] = props["label"]
            for propname, propval in props.items():
                if propname != "label":
                    resdict[propname] = propval
            ut.write_dict(rnafile, resdict)
        ut.show_info("RNA spots saved in file "+str(rnafile))

    def get_drawnspots(self, rnachan, size):
        imgoverlap = None
        rnaspot = self.rnas[rnachan]
        return rnaspot.draw_nonoverlap_spots3D(imgoverlap, self.imshape, size=size)

    def measure_spots(self, rnachan, measurechan=-1, measureimg=None, name=""):
        """ Measure intensity in the spots in the given channel """
        if measurechan >= 0:
            measureimg = self.image[measurechan]
            name = "C"+str(measurechan)
        if measureimg is None:
            ut.show_warning("No image or channel provided for measuring intensity")
            return None
        if name == "":
            name = str("Layer"+str(measurechan))
        self.rnas[rnachan].measure_spots_intensity( intensity_image = measureimg, name=name )
        return self.rnas[rnachan].measures[name]
    
    def threshold_spots_measure( self, rnachan, measure_name, threshold=0.5 ):
        """ Returns the measure if it exists """
        if self.rnas.get(rnachan) is None:
            print("RNA "+str(rnachan)+" not found")
            return None
        if self.rnas[rnachan].measures.get(measure_name) is None:
            print("Measure "+str(measure_name)+" not found for RNA "+str(rnachan))
            return None
        measures = self.rnas[rnachan].measures[measure_name]
        self.rnas[rnachan].measures[measure_name] = np.greater_equal(measures, threshold).astype(float).tolist()


    def get_spots_measure( self, rnachan, measure_name ):
        """ Returns the measure if it exists """
        if self.rnas.get(rnachan) is None:
            print("RNA "+str(rnachan)+" not found")
            return None
        if self.rnas[rnachan].measures.get(measure_name) is None:
            print("Measure "+str(measure_name)+" not found for RNA "+str(rnachan))
            return None
        return self.rnas[rnachan].measures[measure_name]

    def get_spots(self, rnachan):
        return self.rnas[rnachan].get_pointswithprops()

    def set_spots(self, rnachan, spots):
        self.rnas[rnachan].set_points(spots)
    
    def update_spotsAndCountFromPoints(self, spots, labels, scores, chan):
        self.rnas[chan].update_spotsFromPoints(spots, labels, scores, self.pop)
    
    def get_rnalist(self):
        return [i for i in self.rnas.keys()]
    
    def get_done_rnalist(self):
        res = []
        for i, rna in self.rnas.items():
            if len(np.unique(rna.labels)) > 1:
                res.append(i)
        return res

    def get_overlapping_channels( self, channel ):
        """ Find saved overlapping RNASpots that contained the current channel """
        
        ## Check if overlapping spots already in the list of loaded RNAs
        for rnachan in self.rnas.keys():
            if isinstance(rnachan, list):
                ## if the channel has been searched for overlapping points
                if channel in rnachan:
                    return rnachan
        
        ## Otherwise check if a file exists with the correct name containing over RNA
        rnafile = os.path.join(self.resdir, self.imagename+"_RNA_over_")
        saved_files = ut.get_filelist(rnafile+"*.csv")
        for fullfile in saved_files:
            cfile = os.path.basename(fullfile)
            rna_nums = cfile.split("_RNA_over_")
            rna_nums = rna_nums[1]
            rna_nums = rna_nums[0:len(rna_nums)-4]
            for num in list(rna_nums):
                if num.isdigit() and int(num) == channel:
                    self.load_rnafile( fullfile, rna_nums )
                    return rna_nums
        return None

    def find_spot(self, channel, spot, distance=0.2):
        """ Check if RNASpot of channel contains a spot within distance of spot """
        if channel not in self.rnas.keys():
            return False
        return (self.rnas[channel].find_spot( spot, self.scaleXY, self.scaleZ, distance )>= 0 )

    def mixchannels(self, chanlist):
        """ Normalize and add the channels """
        img = np.zeros(self.get_image_shape(in2d=False), np.float16)
        img = img + 1
        #one_max = floor(255/len(chanlist))
        for chan in chanlist:
            norm_image = ut.normalize_img(self.get_channel(chan))
            img = img * norm_image
        #img[img<(one_max*(0.25+len(chanlist)/2.0))] = 0
        #img[img<one_max] = img[img<one_max]*0.25
        img = adjust_gamma(img, 2)
        return img

    def associateCN(self, nucleus, cell):
        self.pop.associateCNLabels(nucleus, cell)
        self.nucmask = self.pop.imgnuc

    def go_association( self, distance, pbar=None ):
        ut.show_info("Start doing association...")
        if pbar is not None:
            pbar.total = 2
        if len(self.pop.nuclei) <= 0:
            self.pop.createNucleiFromMask(associate=False, verbose=self.verbose, scaleXY=self.scaleXY, scaleZ=self.scaleZ)
        if pbar is not None:
            pbar.update(1)
        self.pop.associateCellsNuclei( association_limit=distance, scaleXY=self.scaleXY, scaleZ=self.scaleZ, verbose=self.verbose, pbar=pbar)
        #self.pop.relabelNuclei()
        self.nucmask = self.pop.imgnuc

    def getAverageCellZ(self):
        meanz = self.pop.getMeanCellsZ()
        return int(meanz)

    def measure_nuclear_intensity( self, channel ):
        """ Measure intensity in segmented nuclei """
        img = self.image[channel]
        results = self.pop.measureNuclearStaining( img, channel )
        return results

    def measureCytoplasmic(self, cytochannels, bgrois, zthickness):
        """ Measure background mean intensity (in the ROI) """
        start_time = time.time()
        meanz = bgrois[0][0][0][0]
        until = meanz + self.zdirection*zthickness
        dbg = min(meanz, until)
        ubg = max(meanz, until)
        dbg = max(dbg, 0)
        ubg = min(ubg, self.imshape[0])
        meanbgints = []
        ut.show_info("Measuring cytoplasmic signals in channels: "+str(cytochannels))
        for cyt, bgroi in zip(cytochannels, bgrois):
            bgroi = bgroi[0]
            cmean = np.mean( self.image[cyt, int(dbg):int(ubg+1),int(bgroi[0][1]):int(bgroi[3][1]), int(bgroi[0][2]):int(bgroi[1][2])] )
            meanbgints.append( cmean )

        results = self.pop.measureCellsCytoplasmic( cytoThickness=zthickness, cytochannels=cytochannels, imagesCyto=self.image, meanbgints=meanbgints, scaleXY=self.scaleXY, scaleZ=self.scaleZ )
        print("Cytoplasmic measures in {:.3f}".format((time.time()-start_time)/60)+" min")
        return results

    def drawCytoplasmicMeasure(self, cytochan, results):
        """ Draw intensity image of the cytoplasmic measure in each cell """
        column = "Cyto"+str(cytochan)+"_MeanNormalisedIntensity"
        proj = self.pop.drawProjectedCyto( results, column )
        return proj 

    def fillFeature(self, featname, value):
        self.pop.fill_measure(featname, value)

    def features_filename(self, ifexist=True):
        """ Default filename of features file """
        filename = self.get_filename(endname="_features.csv", ifexist=ifexist)
        return filename

    def image_feature_from_table(self, featname):
        """ Create the classified feature image from current table """
        featimg = np.zeros(self.get_image_shape(in2d=True), np.uint8)
        self.pop.drawFeatureInCells(featimg, featname)
        return featimg


    def loadFeatureFromImage(self, featname):
        endname = "_feat_"+featname+".tif"
        featimg = self.load_image(self.build_filename(endname))
        self.pop.loadFeature(featname, featimg)
        return featimg

    def loadFeaturesTable(self, endname):
        """ Load all features already measured and saved in the given file, previous version with separated features file """
        featfile = os.path.join(self.resdir, self.imagename+endname)
        if os.path.exists(featfile):
            with open(featfile, 'r') as infile:
                csvreader = csv.DictReader(infile)
                for row in csvreader:
                    self.pop.addFeatureValue(row)
                print("Loaded previous features value in the table: "+str(row.keys()))

    def loadFeatureFile(self, featfile):
        """ Load feature(s) from image/table file """
        if not os.path.exists(featfile):
            print("File "+featfile+" not found")
            return
        filename, filext = os.path.splitext(featfile)
        if filext==".tif":
            featname = filename.split("_feat_")[1]
            featimg = self.load_image(featfile)
            self.pop.loadFeature(featname, featimg)
        if filext==".csv":
            self.loadFeatureTable(featfile)

    def loadCytoplasmicTable( self, cytofile ):
        """ Load all cytoplasmic measures already done from the given file """
        if os.path.exists(cytofile):
            with open(cytofile, 'r') as infile:
                csvreader = csv.DictReader( infile )
                for row in csvreader:
                    self.pop.addCytoplasmicValue(row)
                print("Loaded previous cytoplasmic value in the table")
    
    def loadFeatureTable(self, featfile):
        """ Load all features already measured and saved in the given file """
        if os.path.exists(featfile):
            with open(featfile, 'r') as infile:
                csvreader = csv.DictReader(infile)
                for row in csvreader:
                    self.pop.addFeatureValue(row)
                print("Loaded previous features value in the table: "+str(row.keys()))

    def getFeaturesTable(self):
        return self.pop.getFeaturesTable()

    def getFeaturesList(self):
        """ Get the list of current features """
        featlist = self.pop.getFeaturesHeading()
        return featlist

    def getFeatureMax(self, featname):
        """ Returns the max value of featname measures if found, else None """
        return self.pop.getMaxFeature(featname)

    def classifyCells(self, featname, projchan, threshold_mean, threshold_fill):
        """ Perform initial classification based on thresholding """
        projimg = np.mean(self.image[projchan], axis=0)
        return self.pop.classifyCellsFromThreshold( featname, projimg, threshold_mean, threshold_fill )

    def classifyBoundaryCells( self, featname, border=0, boundary=2 ):
        """ Classify cells if they are on edge/boundary of the tissue """
        return self.pop.classifyCellsFromBoundaries( featname, border, boundary )

    def update_cell_feature(self, featname, celllabel, newvalue ):
        self.pop.putFeatureCellValue(featname, celllabel, newvalue)
    
    def change_cell_feature(self, featname, celllabel, newvalue, featimg ):
        """ Update the value of cell and redraw it """
        self.pop.changeFeatureCellValue( featname, celllabel, newvalue, featimg )
    
    def get_feature_value(self, featname, celllabel):
        res = self.pop.getFeatureCellValue(featname, celllabel)
        if res is None:
            return 0
        return res
    
    def tryHull(self):
        return self.pop.drawCellsHull()

    def tryTogether(self):
        return self.pop.tryCell()

    def calculate_zmap(self, projimg, step_size, window_size):
        """ Calculate the zmap of the cells position based on the projection """
        if self.junstain is None:
            self.check_separation()
            if self.junstain is None:
                ut.show_warning("Load junction staining (separated if necessary) before")
                return
        zmap = np.zeros(projimg.shape, "uint8")
        for i, x in enumerate(range(0, projimg.shape[0], step_size)):
            zmap_cur = process_x( x, step=step_size, projimg=projimg, img=self.junstain, winsize=window_size )
            zmap[x:(x+step_size),:] = zmap_cur
        return zmap

    def tryDiffusion( self, nchan ):
        """ Test drawing cells by diffusive walk """
        from skimage.segmentation import random_walker
        labels = self.getJunctionsImage3D()
        cimg = self.image[nchan]
        print("drawing done")
        diffused = random_walker( cimg, labels, beta=100, tol=0.5 )
        return diffused





import os
from magicgui.widgets import TextEdit
from napari.utils.notifications import show_info
import ast
import csv
import configparser

class Configuration:

    def __init__(self, filename, show=True):
        self.filename = filename
        self.parameters = configparser.ConfigParser()
        if self.has_config():
            self.read_parameterfile()
        self.blabla = TextEdit()
        self.blabla.read_only = True
        self.blabla.name = "Blabla"
        self.blabla.value = "Fish&Feats on file "+filename
        if show:
            self.blabla.show()
        self.blablasaved = self.blabla.value
     
    def has_config(self):
        self.hascfg = os.path.exists(self.filename)
        return self.hascfg

    def read_parameterfile(self):
        """ Read the .cfg file that contains the parameters as dict """
        self.parameters.read(self.filename)

    def write_parameterfile(self):
        """ Save the .cfg file that contains the parameters as dict """
        with open(self.filename, "w") as cfgfile:
            self.parameters.write(cfgfile)

    def read_scale(self, mig):
        """ Read the parameters relative to image scalings and load the values """
        if "ImageScalings" in self.parameters:
            paras = self.parameters["ImageScalings"]
            if "scalexy" in paras:
                mig.scaleXY = float(paras["scalexy"])
            if "scaleZ" in paras:
                mig.scaleZ = float(paras["scalez"])
            if "direction" in paras:
                if paras["direction"].find("top high z") >= 0:
                    mig.zdirection = -1
                elif paras["direction"].find("top low z") >= 0:
                    mig.zdirection = 1
                else:
                    try:
                        mig.zdirection = int(float(paras["direction"]))
                    except:
                        mig.zdirection = 1
                        pass
            var = "junction_channel"
            if var in paras:
                if paras[var] == "None":
                    mig.junchan = None
                else:
                    mig.junchan = int(paras[var])
            var = "nuclei_channel"
            if var in paras:
                if paras[var] == "None":
                    mig.nucchan = None
                else:
                    mig.nucchan = int(paras[var])

    def read_junctions(self):
        """ returns the parameters for junction segmentation """
        if "JunctionSeg" in self.parameters:
            paras = self.parameters["JunctionSeg"]
            return paras
        return None 

    def read_parameter_set(self, name):
        """ returns the set of parameters """
        if name in self.parameters:
            return self.parameters[name]
        return None


    def setText(self, text):
        self.blabla.value = text
    
    def addText(self, text):
        self.blabla.value += "\n"+text
    
    def addTmpText(self, text):
        self.blablasaved = self.blabla.value
        self.addText(text)

    def removeTmpText(self):
        self.blabla.value = self.blablasaved

    def addSectionText(self, text):
        self.blabla.value += "\n\n********************************************\n##"+text+"\n"

    def addGroupParameter(self, name):
        """ Add a set of parameter """
        if name not in self.parameters:
            self.parameters[name] = {} 

    def addParameter(self, paratype, paraname, paravalue):
        """ Add one parameter to the set of parameters """
        if paravalue is None:
            (self.parameters[paratype])[paraname] = "None"
        else:
            (self.parameters[paratype])[paraname] = str(paravalue)

    def addTextParameter(self, cat, paraname, paravalue):
        self.blabla.value += "\n%"+cat+"->"+paraname+" = "+str(paravalue)
    
    def saveText(self):
        with open(self.filename, 'w') as f:
            f.write(self.blabla.value)
        show_info("Informations saved in file "+self.filename+"")



from napari.utils.notifications import show_info
import napari

def show_hello_message():
    import logging
    logger = logging.getLogger(__name__)
    logger.setLevel(20)
    show_info('Hello, world!')
    print('Hello, world')
    logger.info("Message")

def start():
    from fish_feats.Naparing import openImage
    openImage()







""" Create the pipeline icons buttons to choose the next step. Adapted from Napari-assistant code: https://github.com/haesleinhuepf/napari-assistant/blob/main/napari_assistant/_gui/_button_grid.py """

from qtpy.QtCore import QSize
from qtpy.QtGui import QIcon, QColor, QBrush
from qtpy.QtWidgets import QListWidget, QListWidgetItem
from pathlib import Path

ICON_ROOT = Path(__file__).parent / "icons"
STYLES = r"""
    QListWidget{
        min-width: 340;
        background: none;
        font-size: 8pt;
        color: #eee;
    }
    QListWidget::item {
        width: 68;
        height: 85;
        margin: 1;
        padding: 4;
    }
    QListWidget::item::hover {
        background: #8A929C;
        width: 68;
        height: 85;
        margin: 1;
        padding: 4;
    }

"""

def _get_background_brush():
    background_color = QColor()
    background_color.setNamedColor("#414851")
    background = QBrush(1)
    background.setColor(background_color)

    return background

def _get_highlight_brush():
    highlight_color = QColor()
    highlight_color.setNamedColor("#68707a")
    highlight = QBrush(1)
    highlight.setColor(highlight_color)

    return highlight


def _get_icon(name):
    path = ICON_ROOT / f'{name.lower().replace(" ", "_")}.png'
    if not path.exists():
        return ""
    return str(path)

class ButtonGrid(QListWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setMovement(self.Static)  # The items cannot be moved by the user.
        self.setViewMode(self.IconMode)  # make items icons
        self.setResizeMode(self.Adjust)  # relayout when view is resized.
        self.setUniformItemSizes(True)  # better performance
        self.setIconSize(QSize(76, 46))
        self.setWordWrap(True)
        self.setStyleSheet(STYLES)
        self.setSpacing(2)
        self.item_mapping = {}


    def addItem(self, label : str, tool_tip : str = None):
        if isinstance(label, QListWidgetItem):
            super().addItem(label)

        item = QListWidgetItem(QIcon(_get_icon(label)), label)
        self.item_mapping[label] = item
        item.setBackground(_get_background_brush())
        
        if tool_tip is not None:
            item.setToolTip(tool_tip)

        super().addItem(item)



from qtpy.QtWidgets import QPushButton, QHBoxLayout, QVBoxLayout, QWidget, QGroupBox, QCheckBox, QSpinBox, QSlider, QDoubleSpinBox, QLabel, QComboBox, QLineEdit, QFileDialog, QTabWidget, QListWidget, QAbstractItemView
from qtpy.QtCore import Qt
import fish_feats.Utils as ut
import fish_feats.FishWidgets as fwid
import os 
import numpy as np
from magicgui.widgets import Table
import time

def unremove(layer):
    ut.show_info("Removing layer locked, throw an error ")
    return

def unmove_selection(layer, selection, coord):
    return

class NapaRNA():
    """ Main interface for RNA options in Napari """

    def __init__(self, viewer, mig, cfg):
        """ Initialize widget """
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg
        
        print("********** Detecting/Counting RNAs ********")
        text = "Choose RNA channel to segment/assign \n"
        text += "RNA can be detected using Big-Fish or loaded from previous file \n"
        text += "Segmented RNA are all non assigned (label=1), assigned RNA are colored by their label = assigned cell \n"
        ut.showOverlayText(viewer, text)
        paras = {}
        for chan in range(self.mig.nbchannels):
            self.viewer.layers["originalChannel"+str(chan)].visible = False
            paras = self.load_RNA_parameters(chan, paras)
    
        rnachannel = self.mig.free_channel()
        defmeth = "Segment with BigFish"
        rnafilename = self.mig.rna_filename(rnachannel, how=".csv", ifexist=True)
        if rnafilename == "":
            rnafilename = self.mig.rna_filename(rnachannel, how=".tif", ifexist=True)
        if rnafilename != "":
            defmeth = "Load segmented file"

        ## add main interface
        self.rna_wid = MainRNA( viewer, mig, cfg, rnachannel, defmeth, rnafilename, paras)
        viewer.window.add_dock_widget( self.rna_wid, name="RNAs" )
    

    def load_RNA_parameters(self, channel, paras):
        """ default or loaded RNA parameters """
        paras["RNA"+str(channel)+"_spotZRadiusNm"] = 1000
        paras["RNA"+str(channel)+"_spotXYRadiusNm"] = 250
        paras["RNA"+str(channel)+"_removeSpotInExtremeZ"] = True
        #paras["RNA"+str(channel)+"_drawing_spot_size"] = 3
        paras["RNA"+str(channel)+"_threshold"] = self.mig.get_rna_threshold(channel)
        paras["RNA"+str(channel)+"_automatic_threshold"] = True
        paras["RNA_spot_disp_size"] = 30
    
        load_paras = self.cfg.read_parameter_set("RNA"+str(channel))
        if load_paras is not None:
            intvars = ["RNA"+str(channel)+"_spotZRadiusNm", "RNA"+str(channel)+"_spotXYRadiusNm", "RNA"+str(channel)+"_drawing_spot_size"]
            for var in intvars:
                if var in load_paras:
                    paras[var] = int(float(load_paras[var]))
            var = "RNA"+str(channel)+"_threshold"
            if var in load_paras:
                paras[var] = float(load_paras[var])
            var = "RNA"+str(channel)+"_automatic_threshold"
            if var in load_paras:
                paras[var] = load_paras[var].strip() == "True"
            var = "RNA"+str(channel)+"_removeSpotInExtremeZ"
            if var in load_paras:
                paras[var] = load_paras[var].strip() == "True"
        load_paras_gen = self.cfg.read_parameter_set("RNAs")
        if load_paras_gen is not None:
            var = "RNA_spot_disp_size"
            if var in load_paras_gen:
                paras[var] = int(float(load_paras_gen[var]))
        return paras


class GetRNA(QWidget):
    """ Get (segment or load) one RNA channel - Interface to choose options """

    def __init__( self, napari_viewer, parameters, nchan, mig, cfg, defmethod, rnafilename, main ):
        super().__init__()
        self.viewer = napari_viewer
        self.mig = mig
        self.cfg = cfg
        self.rnachannel = nchan
        self.paras = parameters
        self.main = main
        #self.index_assign = None
        self.index_edit = None
        self.edition = None
        self.measure_intensity = None
        ## define interface
        layout = QVBoxLayout()
        
        ## Choose rna channel to process
        lay_chan = QHBoxLayout()
        chan_lab = QLabel()
        chan_lab.setText("RNA channel")
        lay_chan.addWidget(chan_lab)
        self.nchan = QSpinBox()
        self.nchan.setMinimum(0)
        self.nchan.setMaximum(self.mig.nbchannels-1)
        self.nchan.setSingleStep(1)
        self.nchan.setValue(self.rnachannel)
        lay_chan.addWidget(self.nchan)
        layout.addLayout(lay_chan)

        ## Choose segmentation/loading method
        self.method = QComboBox()
        layout.addWidget(self.method)
        methods = ["Load segmented file", "Segment with BigFish"]
        for i in range(len(methods)):
            self.method.addItem(methods[i])
        self.method.setCurrentIndex(methods.index(defmethod))

        ## Loading file options
        self.widget_loading = QGroupBox("Loading RNA file")
        loading_layout = QVBoxLayout()
        filelayout = QHBoxLayout()
        file_lab = QLabel()
        file_lab.setText("Load file:")
        filelayout.addWidget(file_lab)
        self.filename = QLineEdit()
        self.filename.setText(rnafilename)
        filelayout.addWidget(self.filename)
        browse = QPushButton("Browse", parent=self)
        filelayout.addWidget(browse)
        browse.clicked.connect(self.load_file)
        loading_layout.addLayout(filelayout)
        self.widget_loading.setLayout(loading_layout)
        layout.addWidget(self.widget_loading)

        ## BigFish options
        self.widget_bigfish = QGroupBox("Segment with BigFish")
        bigfish_layout = QVBoxLayout()
        ## spot z radius
        spotz_line = QHBoxLayout()
        spotz_lab = QLabel()
        spotz_lab.setText("Spot Z radius (nm):")
        spotz_line.addWidget(spotz_lab)
        self.spotz_radius = QLineEdit()
        self.spotz_radius.setText(str(self.paras["RNA"+str(self.rnachannel)+"_spotZRadiusNm"]))
        spotz_line.addWidget(self.spotz_radius)
        bigfish_layout.addLayout(spotz_line)
        ## spot xy radius
        spotxy_line = QHBoxLayout()
        spotxy_lab = QLabel()
        spotxy_lab.setText("Spot X,Y radius (nm):")
        spotxy_line.addWidget(spotxy_lab)
        self.spotxy_radius = QLineEdit()
        self.spotxy_radius.setText(str(self.paras["RNA"+str(self.rnachannel)+"_spotXYRadiusNm"]))
        spotxy_line.addWidget(self.spotxy_radius)
        bigfish_layout.addLayout(spotxy_line)
        ## big-fish threshold 
        threshold_line = QHBoxLayout()
        threshold_lab = QLabel()
        threshold_lab.setText("Threshold:")
        threshold_line.addWidget(threshold_lab)
        self.threshold = QLineEdit()
        self.threshold.setText(str(self.paras["RNA"+str(self.rnachannel)+"_threshold"]))
        threshold_line.addWidget(self.threshold)
        self.automatic_threshold = QCheckBox("Automatic")
        self.automatic_threshold.setChecked(True)
        self.threshold.setEnabled(False)
        self.automatic_threshold.stateChanged.connect(self.autothres_changed)
        threshold_line.addWidget(self.automatic_threshold)
        
        bigfish_layout.addLayout(threshold_line)

        self.remove_extrem = QCheckBox("Remove spots in extremes Z")
        self.remove_extrem.setChecked(False)
        bigfish_layout.addWidget(self.remove_extrem)

        self.widget_bigfish.setLayout(bigfish_layout)
        layout.addWidget(self.widget_bigfish)

        ## Actions when parameters are changed
        self.method.currentIndexChanged.connect(self.method_choice)
        self.nchan.valueChanged.connect(self.channel_choice)
        self.method_choice()
        self.channel_choice()

        btn_go = QPushButton("Get RNA channel", parent=self)
        layout.addWidget(btn_go)
        btn_go.clicked.connect(self.go_rna)
        self.setLayout(layout)

    def autothres_changed(self):
        """ Update interface when automatic threshold option is changed """
        self.threshold.setText(str(self.mig.get_rna_threshold(self.rnachannel)))
        self.threshold.setEnabled(not self.automatic_threshold.isChecked())

    def load_file(self):
        """ Select file to load containing segmented/assigned RNA """
        cdir = os.path.dirname(self.filename.text())
        fileName = QFileDialog.getOpenFileName(self,
    "Choose RNA file", cdir, "RNA Files (*.tif *.csv)")
        self.filename.setText(fileName[0])

    def go_rna(self):
        """ Start processing RNA channel with current selected parameters """

        ## if RNA edition is open, refuse to work
        if self.rnachannel in self.main.edition:
            ut.show_warning( "RNA "+str(self.rnachannel)+" edition is already open, finish and close it before")
            return
        ut.remove_layer( self.viewer, "assignedRNA"+str(self.rnachannel) ) 
        
        self.cfg.addGroupParameter("RNA"+str(self.rnachannel))
        self.cfg.addParameter("RNA"+str(self.rnachannel), "RNA"+str(self.rnachannel)+"_method", self.method.currentText())
        self.cfg.addParameter( "RNA"+str(self.rnachannel), "RNA"+str(self.rnachannel)+"_spotZRadiusNm", int(self.spotz_radius.text()) )
        self.cfg.addParameter( "RNA"+str(self.rnachannel), "RNA"+str(self.rnachannel)+"_spotXYRadiusNm", int(self.spotxy_radius.text()) )
        self.cfg.addParameter( "RNA"+str(self.rnachannel), "RNA"+str(self.rnachannel)+"_threshold", float(self.threshold.text()) )
        self.cfg.addParameter( "RNA"+str(self.rnachannel), "RNA"+str(self.rnachannel)+"_removeSpotInExtremeZ", self.remove_extrem.isChecked() )
        #cfg.addParameter( "RNA"+str(rna_channel), "RNA"+str(rna_channel)+"_drawing_spot_size", drawing_spot_size )
        self.cfg.addParameter( "RNA"+str(self.rnachannel), "RNA"+str(self.rnachannel)+"_automatic_threshold", self.automatic_threshold.isChecked() )
        self.cfg.addGroupParameter("RNAs")
        self.cfg.addParameter( "RNAs", "RNA_spot_disp_size", self.main.spot_disp_size )
        self.cfg.write_parameterfile()
        
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=2, descr="Getting RNA, drawing cells in 3D and segment" )
        if self.method.currentText() == "Segment with BigFish":
            ut.showOverlayText(self.viewer, "Doing segmentation of RNA points in channel "+str(self.rnachannel)+"..." )
            rna_channel = self.rnachannel
            ut.show_layer(self.viewer, rna_channel)
            if ("assignedRNA"+str(rna_channel)) in self.viewer.layers:
                self.viewer.layers.remove("assignedRNA"+str(rna_channel))
                ut.remove_widget(self.viewer, "Save RNA"+str(rna_channel))
            thresh = float(self.threshold.text())
            if thresh <= 0.0:
                thresh = None
            if self.automatic_threshold.isChecked():
                thresh = None
            print("------- Segmenting RNA "+str(rna_channel)+" with big-fish ------ ")
            self.segment_rna(thresh)
            pbar.update(1)
        else:
            self.main.add_cell_contours()
            pbar.update(1)
            print("-------- Loading segmented/assigned RNA "+str(self.rnachannel)+" from file -------")
            self.load_segmentationfile_rna()
        ut.close_progress( self.viewer, pbar )
        ut.show_duration( start_time, "RNAs loaded/segmented in " )

    def channel_choice(self):
        """ Update the parameters in the interface for the selected channel """
        self.rnachannel = self.nchan.value()
        rnafilename = self.mig.rna_filename(self.rnachannel, how=".csv", ifexist=True)
        if rnafilename == "":
            rnafilename = self.mig.rna_filename(self.rnachannel, how=".tif", ifexist=True)
        self.filename.setText(rnafilename)
        self.spotz_radius.setText(str(self.paras["RNA"+str(self.rnachannel)+"_spotZRadiusNm"]))
        self.spotxy_radius.setText(str(self.paras["RNA"+str(self.rnachannel)+"_spotXYRadiusNm"]))
        self.threshold.setText(str(self.paras["RNA"+str(self.rnachannel)+"_threshold"]))
        self.automatic_threshold.setChecked( self.paras["RNA"+str(self.rnachannel)+"_automatic_threshold"] )
        self.remove_extrem.setChecked( self.paras["RNA"+str(self.rnachannel)+"_removeSpotInExtremeZ"] )
        #self.drawing_spot_size.value = paras["RNA"+str(self.rnachannel)+"_drawing_spot_size"]

    def method_choice(self):
        """ Change visibility of parameters according to current selected method """
        self.widget_loading.setVisible(self.method.currentText() == "Load segmented file")
        self.widget_bigfish.setVisible(self.method.currentText() == "Segment with BigFish")


    def load_segmentationfile_rna(self ):
        """ Load segmented RNA file """
        self.mig.load_rnafile(self.filename.text(), self.rnachannel, topop=False)
        self.end_segmented_rna()

    def segment_rna(self, threshold):
        chanel = self.rnachannel
        start_time = time.time()
        self.mig.find_rna( self.rnachannel, int(self.spotz_radius.text()), int(self.spotxy_radius.text()), self.remove_extrem.isChecked(), threshold )
        print("RNA big-fish segmentation finished in {:.3f}".format((time.time()-start_time)/60)+" min")
        self.end_segmented_rna()

    def end_segmented_rna(self):
        """ Create points from segmented spots and call editing/assignement widget """
        nchan = self.rnachannel 
        #ut.removeOverlayText(self.viewer)
        spots, labs, scores = self.mig.get_spots(nchan)
        points = np.array(spots)
        labels = np.array(labs, dtype="int")
        scores = np.array(scores, dtype="float")
        unassigned = (labels==-1) + (labels==1)
        labels[unassigned] = 1
        point_properties = { 'label': labels, 'score':scores, 'unassigned': unassigned, 'intensity': np.array([0.0]*len(labs)) }
        fcolor = []
        if "CellContours" in self.viewer.layers:
            for lab in labels:
                fcolor.append( self.viewer.layers["CellContours"].get_color(lab) )
        else:
            fcolor = "white"

        size = self.main.spot_disp_size
        ut.add_point_layer( self.viewer, points, fcolor, layer_name="assignedRNA"+str(nchan), mig=self.mig, size=size, pts_properties=point_properties ) 
        ut.view_3D( self.viewer )
    
        self.main.go_edition( self.rnachannel )
        self.main.go_measure_intensity( self.rnachannel )
    

    def finish_assign(self):
        """ Close the assign/edit tabs """
        #if self.index_edit is not None:
        #    self.tabs.removeTab(self.index_edit)
        #if self.index_assign is not None:
        #    self.tabs.removeTab(self.index_assign)
        self.edition = None
        self.measure_intensity = None


class PointEditing(QWidget):
    """ Handle RNA assignement + manual correction of RNA segmentation and assignement """

    def __init__(self, chan, viewer, mig, cfg, main):
        super().__init__()
        self.viewer = viewer
        self.channel = chan
        self.mig = mig
        self.cfg = cfg
        self.main = main
        paras = self.cfg.read_parameter_set("RNA"+str(self.channel)+"_Assignement")
        layout = QVBoxLayout()
        disp_settings = {}
        disp_settings["Help button"] = ut.get_color("help")
        disp_settings["auto"] = ut.get_color("group1")
        disp_settings["group2"] = ut.get_color("group2")
        disp_settings["group3"] = ut.get_color("group3")
        ## Assignement options
        auto_line, auto_cbox, self.group_assign = fwid.checkgroup_help( "Automatic assignement", True, "Show/hide automatic assignement options", "Get-rnas", display_settings=disp_settings, groupnb="auto" )
        self.point_editing(paras)
        layout.addLayout( auto_line )
        auto_cbox.setChecked(False)
        layout.addWidget( self.group_assign )
        ## Correction options
        corr_line, group_cbox, self.group_correction = fwid.checkgroup_help( "Point correction", True, "Show/hide RNA correction options", "Get-rnas", display_settings=disp_settings, groupnb="group2" )
        self.interface_corrections()
        layout.addLayout( corr_line )
        group_cbox.setChecked(False)
        layout.addWidget( self.group_correction )
        ## Display options
        disp_line, group_cbox_disp, self.group_display = fwid.checkgroup_help( "Point display", False, "Show.hide options to change points display", "Get-rnas", display_settings=disp_settings, groupnb="group3" )
        self.interface_display()
        layout.addLayout( disp_line )
        layout.addWidget( self.group_display )
        
        ## Save/finish the current RNA
        btn_draw_counts = fwid.add_button( "Draw RNA "+str(self.channel)+" counts", self.draw_rna_counts, "Create a new layer containing the cells colored by their number of assigned RNAs" )
        btn_save_draw = fwid.add_button( "Save drawn RNA "+str(self.channel)+" counts", self.save_draw_rna_counts, "Save the layer CountsRNA"+str(self.channel)+" if it was created" )
        fline = fwid.double_button( btn_draw_counts, btn_save_draw )
        layout.addLayout(fline)
        btn_finish_rna = fwid.add_button( "Save and quit RNA "+str(self.channel)+"", self.finish_rna, "RNA assignement is finished, close this panel", color=ut.get_color("done") )
        layout.addWidget( btn_finish_rna )

        self.setLayout(layout)
        self.active_shortcuts()
        self.show_help_text()
        funcType = type(self.layerrna._move)
        self.layerrna._move = funcType(unmove_selection, self.layerrna)
        #self.viewer.layers.events.removing.connect(lambda event: unremove(self.layerrna))


    def interface_corrections(self):
        """ RNA segmentation/assignement manual correction parameters """
        group_correction_layout = QVBoxLayout()

        ## parameters of cell to assign to
        tocell_line = QHBoxLayout()
        tocell_lab = QLabel()
        tocell_lab.setText("Assign selected points to cell:")
        tocell_line.addWidget(tocell_lab)
        self.assign_tocell = QLineEdit()
        self.assign_tocell.setText("2")
        tocell_line.addWidget(self.assign_tocell)
        group_correction_layout.addLayout(tocell_line)
        sc_lab = QLabel()
        sc_lab.setText( "<Right-click> on a cell to get to its value" )
        group_correction_layout.addWidget(sc_lab)
        ## go assign current points
        btn_assign_selected = QPushButton("Assign selected points", parent=self)
        group_correction_layout.addWidget(btn_assign_selected)
        btn_assign_selected.clicked.connect(self.assign_selected2cell)
        scn_lab = QLabel()
        scn_lab.setText( "<c> to assign selected points" )
        group_correction_layout.addWidget(scn_lab)
        self.group_correction.setLayout( group_correction_layout )

        
    def interface_display(self):
        """ RNA display parameters """
        ## Points display parameters
        disp_layout = QVBoxLayout()

        ## Dislay size of the point
        line = QHBoxLayout()
        ## Add the name of the slider
        lab = QLabel()
        lab.setText( "Display point size" )
        lab.setToolTip( "Change the displayed size of all the RNA points" )
        line.addWidget( lab )
        ## Ranged-value widget
        self.display_size = QSlider( Qt.Horizontal )
        self.display_size.setMinimum( 1 )
        self.display_size.setMaximum( 100 )
        self.display_size.setSingleStep( 1 ) 
        self.display_size.setValue( self.main.spot_disp_size )
        line.addWidget( self.display_size )
        disp_layout.addLayout( line )
        self.display_size.valueChanged.connect( self.change_display_size )

        ## choice of display mode
        displaymode_line = QHBoxLayout()
        displaymode_lab = QLabel("Point size from: ")
        displaymode_line.addWidget(displaymode_lab)
        self.display_mode = QComboBox()
        self.display_mode.setToolTip( "Display only points whose value is within a selected range" )
        displaymode_line.addWidget(self.display_mode)
        disp_layout.addLayout(displaymode_line)
        
        self.display_mode.addItem("None (all same size)")
        
        ## Choose range of scores to display
        self.display_mode.addItem("assignement score")
        self.display_score_layout()
        disp_layout.addWidget(self.score_range_group)
        
        ## Choose range of intensities to display
        self.display_mode.addItem("point intensity")
        self.display_intensity_layout()
        disp_layout.addWidget(self.intensity_range_group)

        ## set visibility of displaying mode interfaces
        self.display_mode.currentIndexChanged.connect(self.display_mode_selected)
        self.display_mode_selected()

        ## reset display: show all the points colored by cell label
        btn_reset = QPushButton("Reset point display", parent=self)
        btn_reset.clicked.connect(self.reset_display)
        disp_layout.addWidget(btn_reset)
        self.group_display.setLayout( disp_layout )

    def display_mode_selected(self):
        """ Set visibility of display mode interfaces from current selection """
        self.score_range_group.setVisible( self.display_mode.currentText() == "assignement score")
        self.intensity_range_group.setVisible( self.display_mode.currentText() == "point intensity")

    def display_intensity_layout(self):
        """ Interface for display point by intensity """
        self.intensity_range_group = QGroupBox("Intensity")
        intensity_layout = QVBoxLayout()
        
        ## Choose channel to measure intensity
        lay_chan = QHBoxLayout()
        chan_lab = QLabel()
        chan_lab.setText("Intensity channel")
        lay_chan.addWidget(chan_lab)
        self.nchan = QSpinBox()
        self.nchan.setMinimum(0)
        self.nchan.setMaximum(self.mig.nbchannels-1)
        self.nchan.setSingleStep(1)
        self.nchan.setValue(self.channel)
        lay_chan.addWidget(self.nchan)
        intensity_layout.addLayout(lay_chan)

        self.nchan.valueChanged.connect(self.measure_point_intensity)

        ## min value
        minint_layout = QHBoxLayout()
        minint_lab = QLabel()
        minint_lab.setText("From:")
        minint_layout.addWidget(minint_lab)
        self.min_intensity = QDoubleSpinBox()
        self.min_intensity.setMinimum(0)
        self.min_intensity.setMaximum(1)
        self.min_intensity.setSingleStep(0.05)
        self.min_intensity.setValue(0)
        minint_layout.addWidget(self.min_intensity)
        intensity_layout.addLayout(minint_layout)
        
        ## max value
        maxintensity_layout = QHBoxLayout()
        maxintensity_lab = QLabel()
        maxintensity_lab.setText("To:")
        maxintensity_layout.addWidget(maxintensity_lab)
        self.max_intensity = QDoubleSpinBox()
        self.max_intensity.setMinimum(0)
        self.max_intensity.setMaximum(1)
        self.max_intensity.setSingleStep(0.05)
        self.max_intensity.setValue(1)
        maxintensity_layout.addWidget(self.max_intensity)
        intensity_layout.addLayout(maxintensity_layout)
        
        self.min_intensity.valueChanged.connect( self.changed_minintensity )
        self.max_intensity.valueChanged.connect( self.changed_maxintensity )
        self.intensity_range_group.setLayout(intensity_layout)


    def display_score_layout(self):
        """ Interface for display point range from score """
        self.score_range_group = QGroupBox("Assignement score")
        score_range_layout = QVBoxLayout() 
        
        ## min value
        minscore_layout = QHBoxLayout()
        minscore_lab = QLabel()
        minscore_lab.setText("From:")
        minscore_layout.addWidget(minscore_lab)
        self.min_score = QDoubleSpinBox()
        self.min_score.setMinimum(0)
        self.min_score.setMaximum(2)
        self.min_score.setSingleStep(0.05)
        self.min_score.setValue(0)
        minscore_layout.addWidget(self.min_score)
        score_range_layout.addLayout(minscore_layout)
        ## max value
        maxscore_layout = QHBoxLayout()
        maxscore_lab = QLabel()
        maxscore_lab.setText("To:")
        maxscore_layout.addWidget(maxscore_lab)
        self.max_score = QDoubleSpinBox()
        self.max_score.setMinimum(0)
        self.max_score.setMaximum(2)
        self.max_score.setSingleStep(0.05)
        self.max_score.setValue(2)
        maxscore_layout.addWidget(self.max_score)
        score_range_layout.addLayout(maxscore_layout)

        self.min_score.valueChanged.connect( self.changed_minscore )
        self.max_score.valueChanged.connect( self.changed_maxscore )
        self.score_range_group.setLayout(score_range_layout)

    def changed_minscore(self, i):
        """ Display min score has been changed """
        ## ensures that max score > min score
        if i > self.max_score.value():
            self.max_score.setValue(min(i+0.05,2))
        ## now update points display from range
        self.change_display_fromscore()

    def changed_maxscore(self, i):
        """ Display max score has been changed """
        ## ensures that max score > min score
        if i < self.min_score.value():
            self.min_score.setValue(max(i-0.05,0))
        ## now update points display from range
        self.change_display_fromscore()

    def change_display_fromscore(self):
        """ Show only points that have a score in the given range """
        self.layerrna.size = 0.1
        ## select corresponding points
        minscore = self.min_score.value()
        maxscore = self.max_score.value()
        self.layerrna.selected_data = {}
        for ind, score in enumerate(self.layerrna.properties['score']):
            if (score >= minscore) and (score <= maxscore):
                self.layerrna.selected_data.add(ind)
        selection = self.layerrna.selected_data
        drawsize = self.layerrna.current_size
        for ind in selection:
            self.layerrna.size[ind] = drawsize
        self.layerrna.selected_data = {}
        self.layerrna.refresh()
    
    def change_display_size(self):
        """ Change the displayed size of all points """
        self.layerrna.size = int( self.display_size.value() )
        self.main.spot_disp_size = int( self.display_size.value() )
        self.layerrna.refresh()

    def measure_point_intensity(self):
        """ Measure the intensity in the points in the current channel """
        nchan = self.nchan.value()
        self.mig.set_spots(self.channel, self.layerrna.data)
        intens = self.mig.measure_spots( self.channel, nchan )
        ## normalize the intensity by its max
        intens = intens/np.max(intens)
        for ind, inte in enumerate(intens):
            self.layerrna.properties['intensity'][ind] = inte
    
    def change_display_fromintensity(self):
        """ Show only points that have intensity in the given range """
        self.measure_point_intensity()
        self.layerrna.size = 0.1
        ## select corresponding points
        minintensity = self.min_intensity.value()
        maxintensity = self.max_intensity.value()
        self.layerrna.selected_data = {}
        for ind, intensity in enumerate(self.layerrna.properties['intensity']):
            if (intensity >= minintensity) and (intensity <= maxintensity):
                self.layerrna.selected_data.add(ind)
        selection = self.layerrna.selected_data
        drawsize = self.layerrna.current_size
        for ind in selection:
            self.layerrna.size[ind] = drawsize
        self.layerrna.selected_data = {}
        self.layerrna.refresh()
    
    def changed_minintensity(self, i):
        """ Display min intensity has been changed """
        ## ensures that max intensity > min intensity
        if i > self.max_intensity.value():
            self.max_intensity.setValue(min(i+0.05,1))
        ## now update points display from range
        self.change_display_fromintensity()
    
    def changed_maxintensity(self, i):
        """ Display max intensity has been changed """
        ## ensures that max intensity > min intensity
        if i < self.min_intensity.value():
            self.min_intensity.setValue(max(i-0.05,0))
        ## now update points display from range
        self.change_display_fromintensity()

    def save_draw_rna_counts( self ):
        """ Save the layer CountRNA if it was created """
        if "CountRNA"+str(self.channel) not in self.viewer.layers:
            ut.show_warning("No count image created yet. Click on Draw RNA counts before" )
            return
        layer = self.viewer.layers["CountRNA"+str(self.channel)]
        #outname = self.mig.build_filename( endname="_RNA"+str(self.channel)+"Counts.png" )
        #actlayer = self.viewer.layers.selection.active.name
        #ut.set_active_layer( self.viewer, "CountRNA"+str(self.channel) )
        #self.viewer.export_figure( outname )
        #ut.set_active_layer( self.viewer, actlayer)
        self.mig.save_image( layer.data, endname="_CountRNA"+str(self.channel)+".tif" )

    
    def draw_rna_counts( self ):
        """ Draw cells with their RNA counts """
        self.main.measure_rna( self.channel )
        countName = self.mig.get_measure_name( self.channel )
        countimg = self.mig.image_count_from_table( countName )
        cmap = "gray"
        if "originalChannel"+str(self.channel) in self.viewer.layers:
            layer = self.viewer.layers["originalChannel"+str(self.channel)]
            cmap = layer.colormap
        ut.remove_layer(self.viewer, "CountRNA"+str(self.channel))
        self.viewer.add_image( countimg, name="CountRNA"+str(self.channel), blending="additive", scale=(self.mig.scaleXY, self.mig.scaleXY), colormap=cmap )


    
    def finish_rna(self):
        """ Assignement and edition of RNA is finished, close everything """
            
        ## save last parameters
        self.cfg.addGroupParameter("RNA"+str(self.channel)+"_Assignement")
        self.cfg.addParameter("RNA"+str(self.channel)+"_Assignement", "assignement_method", self.method.currentText())
        self.cfg.addParameter("RNA"+str(self.channel)+"_Assignement", "limit_distance_to_assign", float(self.limit_distance.text()))
        self.cfg.addParameter("RNA"+str(self.channel)+"_Assignement", "nb_closest_rnas", int(self.nclosest.text()))
        self.cfg.addParameter("RNA"+str(self.channel)+"_Assignement", "assign_when_above", self.keep_above.isChecked())
        self.cfg.addParameter("RNA"+str(self.channel)+"_Assignement", "nb_z_above", int(self.naboveZ.text()))
        self.cfg.write_parameterfile()
            
        self.main.close_edit( self.channel, save=True )

    def save_rnafile(self):
        """ Save current points to RNA file """
        spots = self.layerrna.data
        labels = self.layerrna.properties['label']
        scores = self.layerrna.properties['score']
        filename = self.mig.rna_filename(chan=self.channel, ifexist=False)
        self.mig.save_spots(spots, labels, scores, self.channel, filename)

    def show_help_text(self):
        """ Show help text for RNA editing """
        help_text = "<v> to show/hide cell contours \n"
        help_text += "<f> show/hide the current RNA layer \n"
        help_text += "<d> increase/decrease the RNA opacity \n"
        help_text += "<F1>, <F2>... show/hide the first/second... layer \n"
        help_text = help_text + "<Right-click> on a cell to set the assignement value \n"
        help_text = help_text + "<c> assign selected points to the current assignement value \n"
        help_text += "<u> select all unsassigned points \n"
        help_text += "<s> select points with current assignement value\n"
        help_text += "<a> select all visible points \n"
        help_text += "<o> show only selected points \n"
        help_text += "<r> reset: show all points \n"
        help_text += "<l> show only selected cell \n"
        help_text += "<Control-left click> to select points \n"
        help_text += "<Control-right click> unselect \n"
        help_text += "<Shift-s> shuffle the colors of cells and points \n"
        help_text += "<Alt>+<Left-click> zoom on the clicked position \n"
        help_text += "<Alt>+<Right-click> unzoom \n"
        header = ut.helpHeader(self.viewer, None)
        ut.showOverlayText(self.viewer, header+help_text)
        print( "############ Edit RNA assignement where necessary" )

    def assign_selected2cell(self):
        """ Assign all selected points to the current cell value """
        selection = self.layerrna.selected_data
        assignement = int(self.assign_tocell.text())
        col = self.viewer.layers["CellContours"].get_color( assignement )
        for ind in selection:
            self.layerrna.properties['label'][ind] = assignement 
            self.layerrna.properties['score'][ind] = 2 ## manually assigned = sure
            self.layerrna.properties['unassigned'][ind] = (assignement <= 1) 
            self.layerrna.face_color[ind] = col
        self.layerrna.selected_data = {}
        self.layerrna.refresh()

    def reset_display(self):
        """ Reset points display """
        self.layerrna.size = self.layerrna.current_size
        self.layerrna.refresh()
        self.min_score.setValue(0)
        self.max_score.setValue(2)
        self.min_intensity.setValue(0)
        self.max_intensity.setValue(1)

    def show_hide( self, intlayer ):
        """ Show/hide the ith-layer """
        if intlayer < len( self.viewer.layers ):
            self.viewer.layers[intlayer].visible = not self.viewer.layers[intlayer].visible

    def active_shortcuts(self):
        """ Define shortcuts for point/assignement editing """ 
        @self.layerrna.bind_key('v', overwrite=True)
        def set_label(layerrna):
            if "CellContours" in self.viewer.layers:
                contlay = self.viewer.layers["CellContours"]
                contlay.visible = not contlay.visible
        
        @self.layerrna.bind_key('f', overwrite=True)
        def show_points(layerrna):
            self.layerrna.visible = not self.layerrna.visible
        
        @self.layerrna.bind_key('d', overwrite=True)
        def show_points_opacity(layerrna):
            op = self.layerrna.opacity
            if op >= 0.99:
                self.layerrna.opacity = 0.3
            if op <= 0.5:
                self.layerrna.opacity = 1.0
        
        
        @self.layerrna.bind_key('r', overwrite=True)
        def show_reset(layer):
            self.reset_display()
        
        @self.layerrna.bind_key('l', overwrite=True)
        def show_only_selected_cell(layer):
            contlay = self.viewer.layers["CellContours"]
            contlay.selected_label = int(self.assign_tocell.text())
            contlay.show_selected_label = not contlay.show_selected_label
        
        @self.layerrna.mouse_drag_callbacks.append
        def click(layer, event):
            if event.type == "mouse_press":
                if (event.button == 2) and ("Control" in event.modifiers):
                    self.layerrna.selected_data = {}

                if (event.button == 1) and ("Control" in event.modifiers):
                    modes = ["select", "pan_zoom"]
                    self.layerrna.mode = "select"
                    while event.type == "mouse_press":
                        yield
                    self.layerrna.mode = "pan_zoom"
                    #if self.layerrna.mode in modes:
                    #    self.layerrna.mode = modes[ (modes.index(self.layerrna.mode)+1)%2 ]
                    #else:
                    #    self.layerrna.mode = "pan_zoom"
            if event.type == "mouse_press":
                if (event.button == 2) and (len(event.modifiers) <= 0):
                    # right click pur
                    if "CellContours" not in self.viewer.layers:
                        self.main.add_cell_contours()
                    if "Cells" not in self.viewer.layers:
                        self.main.add_cell_contours()
                    ut.set_active_layer( self.viewer, self.layerrna.name )

                    contlay = self.viewer.layers["Cells"]
                    contlay.visible = True
                    value = contlay.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                    self.assign_tocell.setText(str(value))
                
                if (event.button == 1) and ("Alt" in event.modifiers):
                    ## Zoom on clicked area
                    self.viewer.camera.center = event.position 
                    self.viewer.camera.zoom = self.viewer.camera.zoom * 4
                
                if (event.button == 2) and ("Alt" in event.modifiers):
                    ## Unzoom
                    self.viewer.camera.zoom = self.viewer.camera.zoom *0.5
        
        
        @self.layerrna.bind_key('u', overwrite=True)
        def select_unassigned(layer):
            self.layerrna.selected_data = {}
            for ind, unass in enumerate(self.layerrna.properties['unassigned']):
                if unass:
                    self.layerrna.selected_data.add(ind)
            self.layerrna.refresh()
        
        @self.layerrna.bind_key('s', overwrite=True)
        def select_value(layer):
            self.layerrna.selected_data = {}
            assignement = int(self.assign_tocell.text())
            for ind, lab in enumerate(self.layerrna.properties['label']):
                if lab == assignement:
                    self.layerrna.selected_data.add(ind)
            self.layerrna.refresh()
        
        @self.layerrna.bind_key('o', overwrite=True)
        def show_only_selected(layer):
            self.layerrna.size = 0.1
            selection = self.layerrna.selected_data
            drawsize = self.layerrna.current_size
            for ind in selection:
                self.layerrna.size[ind] = drawsize
            self.layerrna.refresh()
        
        @self.layerrna.bind_key('c', overwrite=True)
        def updateBis(layer):
            self.assign_selected2cell()
        
        @self.layerrna.bind_key('Shift-s', overwrite=True)
        def shuffle_colors(layer):
            contlay = self.viewer.layers["CellContours"]
            contlay.new_colormap()

            fcolor = []
            for ind, lab in enumerate(self.layerrna.properties["label"]):
                fcolor.append( contlay.get_color(lab) )
            self.layerrna.face_color = fcolor
            self.layerrna.refresh()

    def point_editing(self, paras):
        """ RNA assignement, auto and manual correction """
        self.layerrna = self.viewer.layers["assignedRNA"+str(self.channel)]
        saved_layer_data = None
        saved_layer_prop = None
        
        refrnas = self.mig.get_done_rnalist()
        if refrnas is None:
            refrnas = [0]
        if len(refrnas) <= 0:
            refrnas = [0]

        layout = QVBoxLayout()

        ## Assignment method choice
        meth_line = QHBoxLayout()
        meth_lab = QLabel()
        meth_lab.setText("Assignement method:")
        meth_line.addWidget(meth_lab)
        self.method = QComboBox()
        meth_line.addWidget(self.method)
        methods = ["Projection", "ClosestNucleus", "MixProjClosest", "FromNClosest", "Hull"]
        for i in range(len(methods)):
            self.method.addItem(methods[i])
        ## default method 
        if (paras is not None) and ("assignement_method" in paras):
            self.method.setCurrentIndex(methods.index(paras["assignement_method"]))
        layout.addLayout(meth_line)
        ## assignemnet with nclosest specific parameters
        self.parameters_nclosest(paras)
        layout.addWidget(self.group_nclosest)

        ## advanced assignement parameters, common to all methods
        self.show_advanced = QCheckBox("Show advanced parameters")
        self.show_advanced.setChecked(False)
        layout.addWidget(self.show_advanced)
        self.parameters_advanced(paras)
        layout.addWidget(self.group_advanced)
        ## visibility of the panels
        self.show_advanced_parameters()
        self.parameters_assign_visibility()
        self.show_advanced.stateChanged.connect(self.show_advanced_parameters)
        self.method.currentIndexChanged.connect(self.parameters_assign_visibility)

        ## Apply assignement
        btn_apply = QPushButton("Apply assignment", parent=self)
        layout.addWidget(btn_apply)
        btn_apply.clicked.connect(self.assign_rna)
        ## Remove overlapped points
        btn_remover = QPushButton("Remove overlapping points", parent=self)
        layout.addWidget(btn_remover)
        btn_remover.clicked.connect(self.remove_overlapping)
        space = QLabel()
        space.setText("")
        layout.addWidget(space)

        self.group_assign.setLayout(layout)


    def parameters_assign_visibility(self):
        """ Handle assignement method parameters visibility """
        self.group_nclosest.setVisible(self.method.currentText()=="FromNClosest")

    def show_advanced_parameters(self):
        """ Show/Hide advanced parameters layout """
        self.group_advanced.setVisible(self.show_advanced.isChecked())

    def parameters_advanced(self, parameters):
        """ Interface for setting advanced assignement parameters """
        self.group_advanced = QGroupBox("Advanced parameters")
        layout = QVBoxLayout()
        ## Parameter distance limit
        limitdis_line = QHBoxLayout()
        limitdis_lab = QLabel()
        limitdis_lab.setText("Distance limit to assign (µm):")
        limitdis_line.addWidget(limitdis_lab)
        self.limit_distance = QLineEdit()
        assign_dist = 10.0
        if parameters is not None:
            if "limit_distance_to_assign" in parameters:
                assign_dist = parameters["limit_distance_to_assign"]
        self.limit_distance.setText(str(assign_dist))
        limitdis_line.addWidget(self.limit_distance)
        layout.addLayout(limitdis_line)
        ## Parameter assign above cells or not
        self.keep_above = QCheckBox("Assign also RNA above cells")
        assign_when_above = False
        if parameters is not None:
            if "assign_when_above" in parameters:
                assign_when_above = parameters["assign_when_above"].strip() == "True"
        self.keep_above.setChecked(assign_when_above)
        layout.addWidget(self.keep_above)
        ## Parameter nb Z above cells to keep
        nabove_line = QHBoxLayout()
        nabove_lab = QLabel()
        nabove_lab.setText("Keep nb Z above cell:")
        nabove_line.addWidget(nabove_lab)
        self.naboveZ = QLineEdit()
        nb_z_above = 0
        if parameters is not None:
            if "nb_z_above" in parameters:
                nb_z_above = parameters["nb_z_above"]
        self.naboveZ.setText(str(nb_z_above))
        nabove_line.addWidget(self.naboveZ)
        layout.addLayout(nabove_line)
        self.group_advanced.setStyleSheet("QGroupBox { color: #c3c0df; background-color: #54547d; }")
        self.group_advanced.setLayout(layout)

    def parameters_nclosest(self, parameters):
        """ Interface for FromNClosest assignement method """
        self.group_nclosest = QGroupBox("NClosest method")
        layout = QVBoxLayout()
        ## Parameter of number of closest RNAs to consider
        nclos_line = QHBoxLayout()
        nclos_lab = QLabel()
        nb_closest = 10
        if parameters is not None:
            if "nb_closest_rnas" in parameters:
                nb_closest = parameters["nb_closest_rnas"]
        nclos_lab.setText("Nb closest RNAs:")
        nclos_line.addWidget(nclos_lab)
        self.nclosest = QLineEdit()
        self.nclosest.setText(str(nb_closest))
        nclos_line.addWidget(self.nclosest)
        layout.addLayout(nclos_line)
        ## Add reference RNAs channels choice
        refchans_line = QHBoxLayout()
        refchans_lab = QLabel()
        refchans_lab.setText("Reference (assigned) RNAs:")
        refchans_line.addWidget(refchans_lab)
        self.reference_rnas = QListWidget()
        self.reference_rnas.setSelectionMode(QAbstractItemView.ExtendedSelection)
        refrnas = self.mig.get_done_rnalist()
        if refrnas is None:
            refrnas = [0]
        if len(refrnas) <= 0:
            refrnas = [0]
        for ref in refrnas:
            if ref != self.channel:
                self.reference_rnas.addItem(str(ref))
        self.reference_rnas.selectAll()
        refchans_line.addWidget(self.reference_rnas)
        layout.addLayout(refchans_line)
        self.group_nclosest.setLayout(layout)

    def assign_fromclosest( self, method, limDisAsso, assign_even_above, nAbove, nbclose, refrnas, start_time, pbar ):
        """ Assign RNA from same cell as their closest neighbors """
        if assign_even_above:
            above = None
        else:
            above = nAbove
        self.mig.assign_fromclosestrna( self.channel, method, limDisAsso, above, refrnas, nbclose )
        ut.show_duration( start_time, "Spots assigned in " )
        ut.close_progress( self.viewer, pbar )
        self.end_one_rna(method)

    def assign_spots( self, method, limDisAsso, assign_if_above, nAbove, start_time, pbar ):
        """ Assign RNA to cells with the selected method """
        if assign_if_above:
            above = None
        else:
            above = nAbove
        self.mig.assign_rna( self.channel, method, limDisAsso, above )
        pbar.update(1)
        ut.show_duration( start_time, "Spots assigned in " )
        ut.close_progress( self.viewer, pbar )
        self.end_one_rna( method )

    def assign_rna(self):
        """ Automatic assignement of RNA to cell with chosen parameters """
        self.main.add_cell_contours()
        self.mig.set_spots(self.channel, self.layerrna.data)
        self.layerrna.face_color = "white"
        self.layerrna.refresh()
        ut.show_info("assigning spots...")
        start_time = ut.get_time()
        pbar = ut.start_progress( self.viewer, total=2, descr="Assigning spots with "+self.method.currentText() )
        if self.method.currentText() == "FromNClosest":
            reflist = []
            for item in self.reference_rnas.selectedItems():
                reflist.append(int(item.text()))
            self.assign_fromclosest( self.method.currentText(), float(self.limit_distance.text()), self.keep_above.isChecked(), int(self.naboveZ.text()), int(self.nclosest.text()), reflist, start_time, pbar )
        else:
            self.assign_spots( self.method.currentText(), float(self.limit_distance.text()), self.keep_above.isChecked(), int(self.naboveZ.text()), start_time, pbar )

    def end_one_rna(self, method):
        """ Go to manual correction step """
        #ut.removeOverlayText(self.viewer)
        self.correction_rna(method=method)
    
    def correction_rna( self, method ):
        """ Manual correction step, udpate points display """
        #layerrna = self.viewer.layers["assignedRNA"+str(self.channel)]
        self.viewer.layers.selection.select_only(self.layerrna)
        spots, labs, scores = self.mig.get_spots(self.channel)
        labels = np.array(labs, dtype="int")
        unassigned = (labels==-1) + (labels==1)
        labels[unassigned] = 1
        fcolor = []
        for ind, lab in enumerate(labels):
            fcolor.append( self.viewer.layers["CellContours"].get_color(lab) )
            self.layerrna.properties['unassigned'][ind] = unassigned[ind]
            self.layerrna.properties["label"][ind] = lab
            self.layerrna.properties["score"][ind] = scores[ind]
        self.layerrna.face_color = fcolor
        self.layerrna.size = self.main.spot_disp_size
        self.layerrna.refresh()
        
    def remove_overlapping(self):
        """ Remove spots that had been detected as overlapping in several channels """
        channels = self.mig.get_overlapping_channels( channel = self.channel )
        if channels is None:
            ut.show_warning("No overlapping points found. Redo option Overlapping RNA if necessary")
        else:
            nremoved = 0
            self.layerrna.selected_data.clear()
            for ind, pt in enumerate(self.layerrna.data):
                if self.mig.find_spot(channels, pt, distance=0.5):
                    self.layerrna.selected_data.add(ind)
                    nremoved += 1
            self.layerrna.remove_selected()
            self.layerrna.refresh()
            self.layerrna.selected_data.clear()
            ut.show_info(str(nremoved)+" overlapping spots removed")


#### Measure RNA widget

class MainRNA(QWidget):
    """ Main RNA widget, containing the tab interface, save and finish options """

    def __init__( self, napari_viewer, mig, cfg, rnachannel, defmethod, rnafilename, parameters ):
        super().__init__()
        self.viewer = napari_viewer
        self.mig = mig
        self.cfg = cfg
        self.results = None
        self.edition = {}
        self.measure_intensity = {}
        self.spot_disp_size = parameters["RNA_spot_disp_size"] 

        ## GUI parameters
        layout = QVBoxLayout()
        self.tabs = QTabWidget()
        self.tabs.setObjectName("RNAs")
        layout.addWidget(self.tabs)
        ## add save and finish buttons
        btn_save = fwid.add_button( "Save RNAs", self.save_rnas, descr="Save current RNA file and all RNA measures", color=ut.get_color("save") )
        btn_done = fwid.add_button( "RNAS done", None, descr="Close the RNA interface and layers", color=ut.get_color("done") )
        btns_line = fwid.double_button( btn_save, btn_done )
        layout.addLayout( btns_line )


        ## add les onglets        
        wid_get = GetRNA( self.viewer, parameters, rnachannel, self.mig, self.cfg, defmethod, rnafilename, self )
        self.tabs.addTab( wid_get, "Get RNAs" )

        btn_done.clicked.connect( self.rnas_done )

        #show_table = QPushButton("Show measures table")
        #layout.addWidget(show_table)
        #show_table.clicked.connect(self.show_measures)
        
        #save_table = QPushButton("Save measures table")
        #layout.addWidget(save_table)
        #save_table.clicked.connect(self.save_measures)

        ## drawing RNA counts
        #draw_line = QHBoxLayout()
        #draw_cell_counts = QPushButton("Draw cell counts of ")
        #self.rna_todraw = QComboBox()
        #draw_line.addWidget(draw_cell_counts)
        #draw_line.addWidget(self.rna_todraw)
        #layout.addLayout(draw_line)
        #draw_cell_counts.clicked.connect(self.draw_measures)
        #self.update_drawing_rna_choices()
        
        #finish = QPushButton("RNAs done")
        #layout.addWidget(finish)
        #finish.clicked.connect(self.finish_rnas)
        
        self.setLayout(layout)

    def go_edition( self, chanel ):
        """ Launch edition and interface of given chanel """
        self.edition[chanel] = PointEditing( chanel, self.viewer, self.mig, self.cfg, self )
        self.tabs.addTab(self.edition[chanel], "EditRNA"+str(chanel))
        self.tabs.setCurrentWidget( self.edition[chanel] )
    
    def go_measure_intensity( self, chanel ):
        """ Launch measure intensity interface of given chanel """
        self.measure_intensity[chanel] = PointMeasuring( chanel, self.viewer, self.mig, self.cfg, self )
        self.tabs.addTab(self.measure_intensity[chanel], "MeasureIntensity"+str(chanel))

    def add_cell_contours( self ):
        """ Add cell contours and cells layer if not already present """
        if (self.mig.pop is None) or (self.mig.pop.imgcell is None):
            print("No cells found, will not be able to assign.")
            return
        if "CellContours" not in self.viewer.layers:
            celllab = self.viewer.add_labels(self.mig.getJunctionsImage3D(full=False, thick=2), name="CellContours", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY))
            celllab.opacity = 0.6
            celllab.editable = False
        if "Cells" not in self.viewer.layers:
            cells = self.viewer.add_labels(self.mig.getJunctionsImage3D(), name="Cells", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY))
            cells.opacity = 0
            cells.editable = False
    
    
    def close_edit( self, chanel, save=True ):
        """ Remove the edition option of given RNA channel """
        ## remove widget and layer visibility
        if save:
            self.save_rnas()
        self.viewer.layers["originalChannel"+str(chanel)].visible=False
        ut.remove_layer( self.viewer, "assignedRNA"+str(chanel) )
        if chanel in self.edition:
            tabindex = self.tabs.indexOf( self.edition[chanel] )
            self.tabs.removeTab( tabindex )
        if chanel in self.measure_intensity:
            tabindex = self.tabs.indexOf( self.measure_intensity[chanel] )
            self.tabs.removeTab( tabindex )
        del self.edition[chanel]

    
    def save_rnas(self):
        """ Measure cells and RNA counts """
        for rnac in range( self.mig.nbchannels ):
            if ("assignedRNA"+str(rnac)) in self.viewer.layers:
                player = self.viewer.layers["assignedRNA"+str(rnac)]
                rnalabpoints = player.data
                rnaprops = player.properties
                rnalablabels = player.properties['label']
                rnalabscores = player.properties['score']
                self.mig.update_spotsAndCountFromPoints( rnalabpoints, rnalablabels, rnalabscores, rnac )
                filename = self.mig.rna_filename( chan=rnac, ifexist=False )
                self.mig.save_spots( rnalabpoints, rnaprops, rnac, filename )
        self.mig.measure_counts()
        self.mig.save_results()

    def rnas_done( self ):
        """ Finish the option, close all RNA related interface/layer """
        #self.viewer.layers.events.removing.disconnect(unremove)
        for rnac in range(self.mig.nbchannels):
            ut.remove_widget(self.viewer, "Assign RNA"+str(rnac))
            if "assignedRNA"+str(rnac) in self.viewer.layers:
                self.viewer.layers.remove("assignedRNA"+str(rnac))
        ut.remove_widget(self.viewer, "RNAs")
        ut.remove_layer(self.viewer,"CellContours")
        ut.remove_layer(self.viewer,"Cells")
        ut.removeOverlayText(self.viewer)
        self.cfg.removeTmpText()
    
    def measure_rna( self, rnac ):
        """ Measure RNA counts in cells for given chanel """
        if ("assignedRNA"+str(rnac)) in self.viewer.layers:
            player = self.viewer.layers["assignedRNA"+str(rnac)]
            rnalabpoints = player.data
            rnalablabels = player.properties['label']
            rnalabscores = player.properties['score']
            self.mig.update_spotsAndCountFromPoints( rnalabpoints, rnalablabels, rnalabscores, rnac )
        method = self.mig.rnas[rnac].countName
        self.mig.measure_count( rnac, method )

    def measure_rnas(self):
        """ Measure cells and RNA counts """
        rnalist = self.mig.get_rnalist()
        for rnac in rnalist:
            if ("assignedRNA"+str(rnac)) in self.viewer.layers:
                player = self.viewer.layers["assignedRNA"+str(rnac)]
                rnalabpoints = player.data
                rnalablabels = player.properties['label']
                rnalabscores = player.properties['score']
                self.mig.update_spotsAndCountFromPoints( rnalabpoints, rnalablabels, rnalabscores, rnac)
        self.mig.measure_counts()
        self.results = self.mig.get_counts()

    def save_measures(self):
        self.measure_rnas()
        self.update_drawing_rna_choices()
        self.mig.save_results()
    
    def show_measures(self):
        """ Show a table containing all cell measures """
        self.measure_rnas()
        self.update_drawing_rna_choices()
        if (self.results is None) or len(self.results[0].keys())==0:
            ut.show_info("Nothing measured yet")
            return
        tab = {}
        for k in self.results[0].keys():
            tab[k] = []
            for row in self.results:
                if k not in row:
                    tab[k].append(0)
                else:
                    if row[k] is None:
                        tab[k].append(-9999)
                    else:
                        tab[k].append(row[k])
        Table(tab).show()

    def update_drawing_rna_choices(self):
        """ Update the list of RNA choices to be drawn """
        rnachoices = self.mig.get_done_rnalist()
        self.rna_todraw.clear()
        for rna in rnachoices:
            self.rna_todraw.addItem("RNA"+str(rna))

class PointMeasuring(QWidget):
    """ Handle measuring intensity of RNA segmentation """

    def __init__(self, chan, viewer, mig, cfg, main):
        super().__init__()
        self.viewer = viewer
        self.channel = chan
        self.mig = mig
        self.cfg = cfg
        self.main = main
        self.layerrna = self.viewer.layers["assignedRNA"+str(self.channel)]

        layout = QVBoxLayout()
        inside_line, inside_check, self.inside_gr = fwid.checkgroup_help( "Inside segmented nuclei", True, descr="Measure if points are inside segmented nuclei or not", help_link="Get-RNAS#measure-intensity" )
        inside_layout = QVBoxLayout()
        ## title to explain
        inside_lab = fwid.add_label("Which points are inside segmented nuclei")
        inside_layout.addWidget(inside_lab)
        ## measure inside button
        go_inside = fwid.add_button( "Measure points inside nuclei", self.mark_points_inside_nuclei, descr="Measure if points are inside segmented nuclei or not", color=ut.get_color("go") )
        inside_layout.addWidget(go_inside)
        self.inside_gr.setLayout(inside_layout)
        layout.addLayout(inside_line)
        layout.addWidget( self.inside_gr )

        intensity_line, intensity_check, self.intensity_gr = fwid.checkgroup_help( "Measure raw intensity", True, descr="Measure raw intensity of selected channel into the spots", help_link="Get-RNAS#measure-intensity" )
        intensity_layout = QVBoxLayout()
        # choose layer to measure intensity
        line, self.layer_choice = fwid.list_line( "From layer:", descr="Choose opened layer to measure", func=None )
        self.update_layers_list()
        intensity_layout.addLayout(line)
        measure_button = fwid.add_button( "Measure intensity", self.measure_intensity, descr="Measure intensity of selected layer", color=ut.get_color("go") )
        intensity_layout.addWidget(measure_button)
        self.intensity_gr.setLayout(intensity_layout)
        layout.addLayout(intensity_line)
        layout.addWidget( self.intensity_gr )

        ## reset display button
        reset_button = fwid.add_button( "Reset display", self.reset_display, descr="Reset the display of RNA spots" )
        ## update button
        update_list = fwid.add_button( "Update layers list", self.update_layers_list, descr="Update the list of opened layers" )
        res_line = fwid.double_button( reset_button, update_list )
        layout.addLayout(res_line)
        ## run the measure button
        self.setLayout(layout)

    def mark_points_inside_nuclei( self ):
        """ Mark if points are inside segmented nuclei or not """
        if "segmentedNuclei" in self.viewer.layers:
            nuc = self.viewer.layers["segmentedNuclei"].data
        else:
            nuc = self.mig.get_segmented_nuclei()
        if nuc is None:
            ut.show_warning("No segmented nuclei found. Do Get nuclei before to segment them")
            return
        
        nuc = 1*(nuc > 0) ## dont keep the labels
        ## update spots list
        self.mig.set_spots(self.channel, self.layerrna.data)
        
        ## measure spots
        self.mig.measure_spots( self.channel, measureimg = nuc, name="InsideSegmentedNuclei" )
        self.mig.threshold_spots_measure( self.channel, "InsideSegmentedNuclei", threshold=0.5 )
        self.display_by_intensity( "InsideSegmentedNuclei", 0, 2 )

    def reset_display( self ):
        """ Reset point colors to cell label """ 
        self.layerrna.face_color = 'label'
        self.layerrna.refresh()

    def update_layers_list( self ):
        """ Update list of opened layers """
        self.layer_choice.clear()
        layers = self.viewer.layers
        for lay in layers:
            if not lay.name.startswith("assignedRNA"):
                self.layer_choice.addItem(lay.name)

    def measure_intensity( self ):
        """ Measure intensity of selected layer """
        layer_name = self.layer_choice.currentText()
        if layer_name not in self.viewer.layers:
            ut.show_warning("Layer "+layer_name+" not found")
            return
        ## update spots list
        self.mig.set_spots(self.channel, self.layerrna.data)
        ## measure spots
        self.mig.measure_spots( self.channel, measureimg = self.viewer.layers[layer_name].data, name="Int_"+layer_name )
        minint = np.min(self.viewer.layers[layer_name].data)
        maxint = np.max(self.viewer.layers[layer_name].data)
        self.display_by_intensity( "Int_"+layer_name, minint, maxint )

    def display_by_intensity( self, intensity_name, minint, maxint ):
        """ Color points by intensity measure """
        self.layerrna.selected_data = {}
        intensities = self.mig.get_spots_measure( self.channel, intensity_name )
        self.layerrna.features[intensity_name] = intensities
        self.layerrna.features['intensity'] = intensities
        self.layerrna.refresh()
        self.layerrna.face_color = 'intensity'
        minint = min( np.min(intensities), minint*1.25 )
        maxint = max( np.max(intensities), maxint*0.75 )
        self.layerrna.contrast_limits = (minint, maxint)
        self.layerrna.refresh_colors()


class OverlapRNA( QWidget ):
    """ Handle overlap of RNAs between several channels """

    def __init__( self, viewer, mig, cfg ):
        """ GUI to get overlapping RNAs """
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.cfg = cfg

        ## initalize the GUI
        layout = QVBoxLayout()
        ## choose the channels
        chan_line, self.selected_channels = fwid.add_multiple_list("Select channels", "Choose channels to look for overlapping RNAs")
        layout.addLayout(chan_line)
        for chan in range(self.mig.nbchannels):
            self.selected_channels.addItem(str(chan))

        ## size of the spot
        spot_line, self.spot_sigma = fwid.value_line( "Spot sigma", 1.5, descr="blur the spots with given gaussian size" )
        layout.addLayout(spot_line)
        ## threshold to detect spot positive
        threshold_line, self.threshold = fwid.value_line( "Threshold", 0.25, descr="Threshold of overlap to consider it positif")
        layout.addLayout(threshold_line)

        ## show results image
        self.show_mixed_image = fwid.add_check( "Show mixed image", False, descr="Show the mixed image of the selected channels" )
        layout.addWidget(self.show_mixed_image)

        ## button save, go and done
        go_btn = fwid.add_button( "Find overlapping RNAs", self.find_over_rnas, descr="Find RNAs present in all the selected channels", color=ut.get_color("go") )
        layout.addWidget(go_btn)

        save_btn = fwid.add_button( "Save overlapping RNAs", self.save_overlapping_rnas, descr="Save the found overlapping RNAs list", color=ut.get_color("save") )
        done_btn = fwid.add_button( "Done", self.done, descr="Close the overlapping RNAs interface", color=ut.get_color("done") )
        btn_line = fwid.double_button( save_btn, done_btn )
        layout.addLayout(btn_line)

        self.setLayout(layout)

    def save_overlapping_rnas(self):
        """ Save the found overlapping RNAs """
        channels = self.selected_channels.selectedItems()
        channels = [int(c.text()) for c in channels]
        outname = self.mig.get_filename( "_RNA_over_"+str(channels)+".csv" )
        layerrna = ut.get_layer( self.viewer, "OverlapRNA"+str(channels))
        spots = layerrna.data
        labels = [1]*len(spots)
        scores = [0]*len(spots)
        props = { "label": np.array(labels), "score": np.array(scores) }
        self.mig.save_spots(spots.astype(int), props, str(channels), outname)

    def done(self):
        """ Step is done"""
        channels = self.selected_channels.selectedItems() 
        channels = [int(c.text()) for c in channels]
        layerrna = ut.get_layer( self.viewer, "OverlapRNA"+str(channels))
        spots = layerrna.data
        self.mig.set_spots(str(channels), spots)
        ut.remove_layer(self.viewer, "OverlapRNA"+str(channels))
        ut.remove_widget(self.viewer, "Overlapping RNAs")

    def find_over_rnas(self):
        """ Detect RNAs present in all the selected channels """
        channels = self.selected_channels.selectedItems() 
        channels = [int(c.text()) for c in channels]
        mixed = self.mig.mixchannels(channels)
        for lay in self.viewer.layers:
            lay.visible = False
        for chan in channels:
            lay = ut.get_layer(self.viewer, "originalChannel"+str(chan))
            if lay is not None:
                lay.visible = True
        if self.show_mixed_image.isChecked():
            self.viewer.add_image(mixed, name="Mixchannels"+str(channels), scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY), blending="additive")

        spots = self.mig.find_blobs_in_image( mixed, channels, float(self.spot_sigma.text()), float(self.threshold.text()) )
        points = np.array(spots)
        fcolor = "white"
        ut.remove_layer(self.viewer, "OverlapRNA"+str(channels))
        ut.add_point_layer( viewer=self.viewer, pts=points, colors=fcolor, layer_name="OverlapRNA"+str(channels), mig=self.mig, size=7 )